/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector;

import java.util.Collection;
import fr.ufrt.xmlcorrector.exceptions.abstractOperationException;
import fr.ufrt.xmlcorrector.exceptions.addNodeException;
import fr.ufrt.xmlcorrector.exceptions.badAddressException;
import fr.ufrt.xmlcorrector.exceptions.badOperationException;
import fr.ufrt.xmlcorrector.exceptions.headerNodeException;
import fr.ufrt.xmlcorrector.exceptions.noMinimalTreeException;
import fr.ufrt.xmlcorrector.exceptions.noNodeException;
import fr.ufrt.xmlcorrector.exceptions.onlyCharactersException;
import fr.ufrt.xmlcorrector.exceptions.removeNodeException;
import fr.ufrt.xmlcorrector.matrix.Result;
import fr.ufrt.xmlcorrector.schema.TreeSchema;
import fr.ufrt.xmlcorrector.tree.Tree;

/**
 * Class XmlCorrector.
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public abstract class XmlCorrector {

	TreeSchema ts;
	Tree tree;
	char root;
	int threshold;
	int minimal = 0;

	/**
	 * Constructor to use to create an XmlCorrector
	 * 
	 * @param ts
	 *            TreeSchema based on the DTD to use to correct the tree
	 * @param tree
	 *            Tree to correct
	 * @param threshold
	 *            Threshold to use to end the recursivity
	 */
	public XmlCorrector(TreeSchema ts, Tree tree, int threshold) {

		this.ts = ts;
		this.tree = new Tree(tree);
		this.threshold = threshold;
		root = this.ts.getRoot();

	}

	/**
	 * Constructor used recursivly to correct a sub tree with a given caracter
	 * 
	 * @param ts
	 *            TreeSchema based on the DTD to use to correct the tree
	 * @param tree
	 *            Tree to correct
	 * @param threshold
	 *            Threshold to use to end the recursivity
	 * @param r
	 *            Caracter to use as root to correct
	 */
	protected XmlCorrector(TreeSchema ts, Tree tree, int threshold, char r) {

		this.ts = ts;
		this.tree = new Tree(tree);
		this.threshold = threshold;
		this.root = r;

	}

	/**
	 * Calculates every minimal correction in the given threshold. Corrections
	 * that aren't minimal won't be calculated
	 * 
	 * @return a Result that contains every OperationSequence correcting the
	 *         tree with the minimal cost in the given Threshold. Returns -1 as
	 *         cost if there isn't any correction.
	 * @throws headerNodeException
	 * @throws onlyCharactersException
	 * @throws badAddressException
	 * @throws noNodeException
	 * @throws noMinimalTreeException
	 * @throws badOperationException
	 * @throws addNodeException
	 * @throws removeNodeException
	 * @throws abstractOperationException
	 */
	public abstract Result getMinimalCorrections() throws headerNodeException,
			onlyCharactersException, badAddressException, noNodeException,
			noMinimalTreeException, badOperationException, addNodeException,
			removeNodeException, abstractOperationException;

	/**
	 * Calculates every correction in the given threshold.
	 * 
	 * @return Collection of Result, foreach cost, one Result.
	 * @throws headerNodeException
	 * @throws onlyCharactersException
	 * @throws badAddressException
	 * @throws noNodeException
	 * @throws noMinimalTreeException
	 * @throws badOperationException
	 * @throws addNodeException
	 * @throws removeNodeException
	 * @throws abstractOperationException
	 */
	public abstract Collection<Result> getCorrections()
			throws headerNodeException, onlyCharactersException,
			badAddressException, noNodeException, noMinimalTreeException,
			badOperationException, addNodeException, removeNodeException,
			abstractOperationException;

}
