/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;

import dk.brics.automaton.Automaton;
import dk.brics.automaton.State;
import dk.brics.automaton.Transition;

import fr.ufrt.xmlcorrector.couple.MPair;
import fr.ufrt.xmlcorrector.exceptions.abstractOperationException;
import fr.ufrt.xmlcorrector.exceptions.addNodeException;
import fr.ufrt.xmlcorrector.exceptions.badAddressException;
import fr.ufrt.xmlcorrector.exceptions.badOperationException;
import fr.ufrt.xmlcorrector.exceptions.headerNodeException;
import fr.ufrt.xmlcorrector.exceptions.noMinimalTreeException;
import fr.ufrt.xmlcorrector.exceptions.noNodeException;
import fr.ufrt.xmlcorrector.exceptions.onlyCharactersException;
import fr.ufrt.xmlcorrector.exceptions.removeNodeException;
import fr.ufrt.xmlcorrector.matrix.Cell;
import fr.ufrt.xmlcorrector.matrix.Cells;
import fr.ufrt.xmlcorrector.matrix.Matrix;
import fr.ufrt.xmlcorrector.matrix.Result;
import fr.ufrt.xmlcorrector.matrix.RightExtendedMatrix;
import fr.ufrt.xmlcorrector.matrix.SimpleMatrix;
import fr.ufrt.xmlcorrector.matrix.TreeEditionMatrix;
import fr.ufrt.xmlcorrector.operations.Operation;
import fr.ufrt.xmlcorrector.operations.OperationSequence;
import fr.ufrt.xmlcorrector.operations.edit.AddNodeOperation;
import fr.ufrt.xmlcorrector.operations.edit.RelabelNodeOperation;
import fr.ufrt.xmlcorrector.operations.update.RemoveTreeOperation;
//import fr.ufrt.xmlcorrector.operations.update.ReplaceTreeOperation;
import fr.ufrt.xmlcorrector.schema.TreeSchema;
import fr.ufrt.xmlcorrector.tree.Address;
import fr.ufrt.xmlcorrector.tree.Tree;

/**
 * Class XmlCorrectorMatrix: for computing the tree-to-language-edtion matrix.
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class XmlCorrectorMatrix extends XmlCorrector {

	Hashtable<Integer, Result> results;
	static Hashtable<MPair<Character, Tree>, Hashtable<Integer, Collection<Result>>> sub_trees_corrections = new Hashtable<MPair<Character, Tree>, Hashtable<Integer, Collection<Result>>>();
	Result result;

	// DEBUG: static int nb_tabs = 0;

	// DEBUG: private static String getTabs(int nb) {

	// DEBUG: String str = "";

	// DEBUG: for(int i = 0; i < nb; i++) str+= "\t";

	// DEBUG: return str;

	// DEBUG: }

	public XmlCorrectorMatrix(TreeSchema ts, Tree tree, int threshold) {

		super(ts, tree, threshold);

	}

	private XmlCorrectorMatrix(TreeSchema ts, Tree tree, int threshold, char r) {

		super(ts, tree, threshold, r);

	}

	@Override
	public Collection<Result> getCorrections() throws headerNodeException,
			onlyCharactersException, badAddressException, noNodeException,
			noMinimalTreeException, badOperationException, addNodeException,
			removeNodeException, abstractOperationException {

		// DEBUG:
		// System.out.println(getTabs(nb_tabs)+"<Correction of "+this.tree.toFlatString()+" with "+this.root+">");
		// DEBUG: nb_tabs++;

		// Corrections haven't been corrected yet, or have been but only
		// minimal...

		if (this.results == null || this.minimal != -2) {

			this.minimal = -2;
			this.correction();

			if (this.results.size() == 0)
				this.results.put(-1, new Result(-1));
		}

		// DEBUG: System.out.println(getTabs(nb_tabs)+this.result);
		// DEBUG: nb_tabs--;

		return this.results.values();

	}

	@Override
	public Result getMinimalCorrections() throws headerNodeException,
			onlyCharactersException, badAddressException, noNodeException,
			noMinimalTreeException, badOperationException, addNodeException,
			removeNodeException, abstractOperationException {

		// DEBUG:
		// System.out.println(getTabs(nb_tabs)+"<Correction of "+this.tree.toFlatString()+" with "+this.root+">");
		// DEBUG: nb_tabs++;

		// Minimal corrections haven't been calculated yet

		if (this.result == null) {

			// There is no correction in data, let's calculate them... minimal

			if (this.results == null) {

				// this.minimal will be used during the complete recursion to
				// keep only minimal solutions...
				// if this.minimal >= -1, then we keep only minimal
				// if this.minimal == -2, then we keep everything
				this.minimal = -1;
				this.correction();

			}

			// Lets search for the minimal cost

			int minimal_cost = -1;
			Iterator<Result> iterator = this.results.values().iterator();

			while (iterator.hasNext()) {

				int temp_cost = iterator.next().getCost();
				if (temp_cost < minimal_cost || minimal_cost == -1)
					minimal_cost = temp_cost;

			}

			// Let's create the result
			this.result = new Result(minimal_cost);

			// Let's take only minimal_costs... if there is solution !

			if (minimal_cost >= 0) {

				// Foreach

				iterator = this.results.values().iterator();

				while (iterator.hasNext()) {

					Result r = iterator.next();

					// It's minimal

					if (r.getCost() == minimal_cost) {

						this.result.addSequences(r.getSequences());

					}

					// We just searched every minimal, so no need to keep the
					// not minimal...

					else if (this.minimal != -2) {

						this.results.remove(r.getCost());

					}

				}

			}

		}

		// DEBUG: System.out.println(getTabs(nb_tabs)+this.result);
		// DEBUG: nb_tabs--;

		// DEBUG:
		// System.out.println(getTabs(nb_tabs)+"</Correction of "+this.tree.toFlatString()+" with "+this.root+">");

		return this.result;

	}

	/**
	 * Function that launch the algorithm and recursivity
	 * 
	 * @throws headerNodeException
	 * @throws onlyCharactersException
	 * @throws badAddressException
	 * @throws noNodeException
	 * @throws badOperationException
	 * @throws noMinimalTreeException
	 * @throws addNodeException
	 * @throws removeNodeException
	 * @throws abstractOperationException
	 */
	private void correction() throws headerNodeException,
			onlyCharactersException, badAddressException, noNodeException,
			badOperationException, noMinimalTreeException, addNodeException,
			removeNodeException, abstractOperationException {

		this.results = new Hashtable<Integer, Result>();

		// Do we have to correct it ?

		// If all corrections are looked for (not only minimal), we perform the
		// correction event if the tree is valid
		boolean valide = (this.root == this.ts.getRoot()) ? // Check if the
															// schema root is
															// equal to the
															// target root
		// Check is the tree to correct is valid
		this.minimal != -2 && this.ts.isValid(this.tree)
				:
				// Check if the tree to correct is non empty, if its root is
				// equal to the target root, and if the tree (or subtree) is
				// valid wrt the schema
				this.minimal != -2 && !this.tree.isEmpty()
						&& this.tree.getRoot().getHeader() == this.root
						&& this.ts.isDeepValid(this.tree.getRoot());

		if (!valide) {

			// Has he an automaton ? We can correct him so "Grave " +
			// minimal_cost + " Minimal " + this.minimal + " Seuil " +
			// this.threshold

			if (ts.getAutomaton(this.root) != null) {
				// Initialization of the matrix and compute the first column
				// Size of the matrix
				int width = 1;
				int height = (this.tree.isEmpty()) ? 1 : this.tree.getRoot()
						.getChildrenNumber() + 1;

				Matrix<Cells> matrix = new SimpleMatrix<Cells>(width, height);

				Cell c = null;
				Tree reference = new Tree(this.root);

				if (this.tree.isEmpty() && reference.isEmpty()) {

					c = new Cell(0);

				}

				else if (this.tree.isEmpty() || reference.isEmpty()) {

					OperationSequence os = new OperationSequence();

					if (this.tree.isEmpty()) {

						os.addOperation(new AddNodeOperation(new Address(""),
								reference.getRoot().getHeader()));

						c = new Cell(AddNodeOperation.cost);
						c.addSequence(os);

					} else {

						Operation o = new RemoveTreeOperation(new Address(""));

						os.addOperation(o);

						c = new Cell(o.getCost(this.tree));
						c.addSequence(os);

						for (int i = 0; i < matrix.getHeight(); i++)
							matrix.setValue(0, i, new Cells(c));

						// return new Result(c);

					}

				}

				else if (this.tree.getRoot().getHeader() != reference.getRoot()
						.getHeader()) {

					OperationSequence os = new OperationSequence();
					os.addOperation(new RelabelNodeOperation(new Address(""),
							reference.getRoot().getHeader()));

					c = new Cell(RelabelNodeOperation.cost);
					c.addSequence(os);

				}

				else {

					c = new Cell(0);

					OperationSequence os = new OperationSequence();
					c.addSequence(os);

				}

				// We add the first cell
				matrix.setValue(0, 0, new Cells(c));
				// Let's compute the first column
				for (int i = 1; i < matrix.getHeight(); i++)
					fillMatrix(0, i, matrix, this.root);

				// Let's start recursivity

				Automaton a = ts.getAutomaton(this.root);

				correctionState(matrix, a.getInitialState());

			}

			// He hasn't, let's do a simple compare

			else {

				Tree temp = new Tree(this.root);

				TreeEditionMatrix tem = new TreeEditionMatrix(this.tree, temp,
						this.threshold);

				Result tem_result = tem.compare();

				if (tem_result != null) {

					this.results.put(tem_result.getCost(), tem_result);

					// if we are calculating minimal, we have to change
					// this.minimal to save the fact that it is now minimal
					if (this.minimal != -2) {
						if (this.minimal == -1
								|| (this.minimal >= 0 && tem_result.getCost() < this.minimal))
							this.minimal = tem_result.getCost();
					}

				}

			}

		}

		// We don't have to correct it

		else {

			Result r = new Result(0);
			r.addSequence(new OperationSequence());

			this.results.put(0, r);

		}

	}

	/**
	 * Function that manage a state by forking the calculation for every
	 * transition of the state... If we are in a final state and that the
	 * correction is valid (threshold, minimal, etc), we keep it.
	 * 
	 * @param current_matrix
	 *            Matrix that contains cost and OperationSequence to get to this
	 *            state
	 * @param current_state
	 *            State we're in
	 * @throws badAddressException
	 * @throws noNodeException
	 * @throws headerNodeException
	 * @throws onlyCharactersException
	 * @throws noMinimalTreeException
	 * @throws badOperationException
	 * @throws addNodeException
	 * @throws removeNodeException
	 * @throws abstractOperationException
	 */
	private void correctionState(Matrix<Cells> current_matrix,
			State current_state) throws badAddressException, noNodeException,
			headerNodeException, onlyCharactersException,
			noMinimalTreeException, badOperationException, addNodeException,
			removeNodeException, abstractOperationException {

		// We're in a final state, we have to save the solution ?

		if (current_state.isAccept()) {

			// Get the solution

			Cells solution = current_matrix.getValue(
					current_matrix.getWidth() - 1,
					current_matrix.getHeight() - 1);

			// Is the solution in the threshold, are we going on minimal only ?

			if (solution != null
					&& solution.getMinimalCost() <= this.threshold
					&& (this.minimal <= -1 || this.minimal >= solution
							.getMinimalCost())) {

				// New minimal cost ? Reset all the others
				// if we are not minimal (this.minimal == -2) we won't do it...
				// so we won't change this.minimal

				if (this.minimal == -1
						|| (this.minimal >= 0 && solution.getMinimalCost() < this.minimal)) {

					this.results = new Hashtable<Integer, Result>();
					this.minimal = solution.getMinimalCost();

				}

				// Let's put the solutions
				ArrayList<Cell> cells = solution.getCells();
				for (int i = 0; i < cells.size(); i++) {
					if (this.results.containsKey(cells.get(i).getCost())) {

						this.results.get(cells.get(i).getCost()).addSequences(
								cells.get(i).getSequences());

					} else {

						Result r = new Result(cells.get(i));

						// System.err.println("ajout d'une solution � "+r.getCost()+" < "+this.threshold);

						this.results.put(r.getCost(), r);

					}
				}

			}

		}

		// Now let's do every next transition

		Iterator<Transition> i = current_state.getTransitions().iterator();
		Transition transition;
		char c, c_min, c_max;

		while (i.hasNext()) {

			transition = i.next();
			c_min = transition.getMin();
			c_max = transition.getMax();

			// A transition can accept multiple chars...

			for (c = c_min; c <= c_max; c++) {

				// DEBUG:
				// System.out.println(getTabs(nb_tabs)+"<correction avec "+c+">");
				// DEBUG: nb_tabs++;
				// if(this.minimal == -2 || c != this.root)
				correctionTransition(current_matrix, transition.getDest(), c);
				// DEBUG: nb_tabs--;
				// DEBUG:
				// System.out.println(getTabs(nb_tabs)+"</correction avec "+c+">");

			}

		}

	}

	/**
	 * Function adding a column to the Matrix given, using the given char to
	 * correct, and continuing the algorithm with the given state. Makes a fork
	 * for every Tree candidate.
	 * 
	 * @param current_matrix
	 *            Matrix of the correction
	 * @param next_state
	 *            State to use next
	 * @param c
	 *            character to use to correct
	 * @throws badAddressException
	 * @throws noNodeException
	 * @throws headerNodeException
	 * @throws onlyCharactersException
	 * @throws noMinimalTreeException
	 * @throws badOperationException
	 * @throws addNodeException
	 * @throws removeNodeException
	 * @throws abstractOperationException
	 */
	private void correctionTransition(Matrix<Cells> current_matrix,
			State next_state, char c) throws badAddressException,
			noNodeException, headerNodeException, onlyCharactersException,
			noMinimalTreeException, badOperationException, addNodeException,
			removeNodeException, abstractOperationException {

		int x = current_matrix.getWidth();

		// Let's get a bigger matrix

		Matrix<Cells> temp_matrix = new RightExtendedMatrix<Cells>(
				current_matrix);

		// And fill the matrix

		int j = 0, temp_cost, temp_minimal_collumn_cost = Integer.MAX_VALUE;

		while (j < temp_matrix.getHeight()) {
			temp_cost = fillMatrix(x, j, temp_matrix, c);
			if (temp_cost < temp_minimal_collumn_cost)
				temp_minimal_collumn_cost = temp_cost;
			j++;

		}

		// If it's a ok collumn, let's continue
		// If we are minimal, don't forget to cut the path if it's already more
		// expensive...

		if (temp_minimal_collumn_cost <= this.threshold
				&& (this.minimal <= -1 || temp_minimal_collumn_cost <= this.minimal)) {

			correctionState(temp_matrix, next_state);

		}

	}

	/**
	 * Function that fills the Cell at (x,y) from the given matrix by using the
	 * given root for a schema
	 * 
	 * @param x
	 *            Abscissa where to put the cost and OperationSequence
	 * @param y
	 *            ordinate where to put the cost and OperationSequence
	 * @param matrix
	 *            Matrix where to put the cost and OperationSequence
	 * @param char c to use to do comparisons between the tree and the tree
	 *        schema with c as root
	 * @return cost of the correction in the cell
	 * @throws badAddressException
	 * @throws noNodeException
	 * @throws headerNodeException
	 * @throws badOperationException
	 * @throws abstractOperationException
	 * @throws removeNodeException
	 * @throws addNodeException
	 * @throws noMinimalTreeException
	 * @throws onlyCharactersException
	 */
	private int fillMatrix(int x, int y, Matrix<Cells> matrix, char c)
			throws badAddressException, noNodeException, headerNodeException,
			badOperationException, onlyCharactersException,
			noMinimalTreeException, addNodeException, removeNodeException,
			abstractOperationException {
		int cost;
		int minimal_cost = Integer.MAX_VALUE;
		Cells cell;

		HashMap<Integer, HashSet<OperationSequence>> minimal_os = new HashMap<Integer, HashSet<OperationSequence>>();
		HashSet<OperationSequence> temp_al_os;
		Iterator<OperationSequence> hash_iterator;

		Address address;

		// First: Look upside !
		// It's a RemoveTreeOperation

		if (matrix.getValue(x, y - 1) != null) {

			// First get the cell
			cell = matrix.getValue(x, y - 1);
			if (cell.getMinimalCost() < this.threshold
					&& (this.minimal <= -1 || cell.getMinimalCost() < this.minimal)) {
				// The address of the Node to add
				address = new Address(String.valueOf(y - 1));

				ArrayList<Cell> thecells = cell.getCells();

				for (int i = 0; i < thecells.size(); i++) {
					// Get the total cost...
					// System.out.println(this.tree);
					cost = thecells.get(i).getCost()
							+ this.tree.getNodeAsTree(address).getRemovalCost();

					// if it is minimal, we have to save the sequence...
					if (cost <= this.threshold
							&& (this.minimal <= -1 || cost <= this.minimal)) {

						// If the cost is minimal, let's save it
						if (cost < minimal_cost)
							minimal_cost = cost;

						// let's copy every sequence of the previous cell
						temp_al_os = new HashSet<OperationSequence>(thecells
								.get(i).getSequences());
						hash_iterator = temp_al_os.iterator();

						// We need a var
						OperationSequence temp_os;

						// For every sequence, let's add our tree...
						while (hash_iterator.hasNext()) {

							temp_os = hash_iterator.next();

							// let's create the address with the offset...
							Address address2 = new Address(String.valueOf(y
									- 1
									+ temp_os.getAddressOffset(new Address(
											String.valueOf(y - 1)))));

							// and add it to the operations :/
							temp_os.addOperation(new RemoveTreeOperation(
									address2));

						}

						if (minimal_os.containsKey(cost))
							minimal_os.get(cost).addAll(temp_al_os);
						else
							minimal_os.put(cost, temp_al_os);

					}

				}
			}

		}

		// Second: Look leftside !
		// It's an AddTreeOperation
		/*
		 * if(matrix.getValue(x - 1, y) != null) {
		 * 
		 * // first get the cell cell = matrix.getValue(x - 1, y);
		 * if(cell.getMinimalCost() < this.threshold && (this.minimal <= -1 ||
		 * cell.getMinimalCost() < this.minimal)) {
		 * 
		 * if(this.ts.getMinimalTrees(c) != null) { //we get the minimal tree of
		 * the tree schema which root is c if mininal is true //or all trees
		 * under the threshold if minimal is false ArrayList<Tree> minimal_trees
		 * = this.ts.getMinimalTrees(c);
		 * 
		 * // then the address of the node to remove address = new
		 * Address(String.valueOf(x - 1));
		 * 
		 * ArrayList<Cell> thecells = cell.getCells(); for(int i=0; i<
		 * thecells.size(); i++ ) { for(int j=0; j<minimal_trees.size(); j++) {
		 * Tree minimal_tree = minimal_trees.get(j);
		 * 
		 * // The minimal operation AddTreeOperation addTreeOperation = new
		 * AddTreeOperation(address, minimal_tree);
		 * 
		 * // the cost cost = thecells.get(i).getCost() +
		 * addTreeOperation.getCost(minimal_tree);
		 * 
		 * // then we should save results if(cost <= this.threshold &&
		 * (this.minimal <= -1 || cost <= this.minimal)) {
		 * 
		 * // if the cost is minimal... if(cost < minimal_cost) minimal_cost =
		 * cost;
		 * 
		 * // first copy the OperationSequences of the left cell temp_al_os =
		 * new HashSet<OperationSequence>(thecells.get(i).getSequences());
		 * hash_iterator = temp_al_os.iterator();
		 * 
		 * // We need a var OperationSequence temp_os;
		 * 
		 * // On every of the cell, add the Operation
		 * while(hash_iterator.hasNext()) {
		 * 
		 * temp_os = hash_iterator.next();
		 * 
		 * // let's add the operation temp_os.addOperation(addTreeOperation);
		 * 
		 * }
		 * 
		 * if(minimal_os.containsKey(cost))
		 * minimal_os.get(cost).addAll(temp_al_os); else minimal_os.put(cost,
		 * temp_al_os);
		 * 
		 * } }
		 * 
		 * }
		 * 
		 * }
		 * 
		 * }
		 * 
		 * }
		 */

		if (matrix.getValue(x - 1, y) != null) {

			// First we get the Cell
			cell = matrix.getValue(x - 1, y);
			if (cell.getMinimalCost() < this.threshold
					&& (this.minimal <= -1 || cell.getMinimalCost() < this.minimal)) {

				Tree aux = new Tree();
				Collection<Result> temp = subTreeCorrectionGeneration(c, aux,
						this.threshold - cell.getMinimalCost());

				// If the tree can be corrected...
				if (temp != null && temp.size() > 0) {

					ArrayList<Cell> thecells = cell.getCells();
					for (int i = 0; i < thecells.size(); i++) {

						Iterator<Result> ir = temp.iterator();
						while (ir.hasNext()) {
							Result r = ir.next();

							if (r.getCost() != -1) {
								// Getting the costs
								cost = r.getCost() + thecells.get(i).getCost();

								// If it is minimal, we have to add the result ?
								// :)
								// It depends, some correction may be
								// unnecessary
								if (cost <= this.threshold
										&& (this.minimal <= -1 || cost <= this.minimal)) {

									// If the cost is minimal...
									if (cost < minimal_cost)
										minimal_cost = cost;

									// Creating the address
									// address = new Address(String.valueOf(y -
									// 1));
									// Modify the operation sequence
									// r.addAddressPrefix(address);
									temp_al_os = new HashSet<OperationSequence>();
									Iterator<OperationSequence> cellos_iterator = thecells
											.get(i).getSequences().iterator();

									while (cellos_iterator.hasNext()) {
										OperationSequence cellos = cellos_iterator
												.next();

										Iterator<OperationSequence> ros_iterator = r
												.getSequences().iterator();
										while (ros_iterator.hasNext()) {
											OperationSequence ros = ros_iterator
													.next();

											address = new Address(
													String.valueOf(y
															+ cellos.getAddressOffset(new Address(
																	String.valueOf(y)))));
											ros.addAddressPrefix(address);

											OperationSequence cellos2 = new OperationSequence(
													cellos);

											cellos2.addOperationSequence(ros);
											temp_al_os.add(cellos2);

										}
									}

									if (minimal_os.containsKey(cost))
										minimal_os.get(cost).addAll(temp_al_os);
									else
										minimal_os.put(cost, temp_al_os);

								}
							}
						}

					}

				}

			}

		}

		// Third, up-left side !
		// It's a CorrectTreeOperation

		if (matrix.getValue(x - 1, y - 1) != null) {

			// First we get the Cell
			cell = matrix.getValue(x - 1, y - 1);
			if (cell.getMinimalCost() <= this.threshold
					&& (this.minimal <= -1 || cell.getMinimalCost() <= this.minimal)) {
				// Then we create a recursive call correcting the tree
				// DEBUG : System.out.println("Correction de " +
				// this.tree.getNodeAsTree(new Address(String.valueOf(y - 1))) +
				// "par rapport à " + c);
				// And get the result :)

				/*
				 * Collection<Result> temp = new ArrayList<Result>();
				 * if(this.minimal == -2) temp = subTreeCorrectionGeneration(c,
				 * this.tree.getNodeAsTree(new Address(String.valueOf(y - 1))),
				 * this.threshold - cell.getMinimalCost()); else {
				 * XmlCorrectorMatrix correction = new
				 * XmlCorrectorMatrix(this.ts, this.tree.getNodeAsTree(new
				 * Address(String.valueOf(y - 1))), this.threshold -
				 * cell.getMinimalCost() , c);
				 * temp.add(correction.getMinimalCorrections()); }
				 */

				Collection<Result> temp = subTreeCorrectionGeneration(c,
						this.tree.getNodeAsTree(new Address(String
								.valueOf(y - 1))),
						this.threshold - cell.getMinimalCost());

				// If the tree can be corrected...
				if (temp != null && temp.size() > 0) {

					ArrayList<Cell> thecells = cell.getCells();
					for (int i = 0; i < thecells.size(); i++) {

						Iterator<Result> ir = temp.iterator();
						while (ir.hasNext()) {
							Result r = ir.next();

							if (r.getCost() != -1) {
								// Getting the costs
								cost = r.getCost() + thecells.get(i).getCost();

								// If it is minimal, we have to add the result ?
								// :)
								// It depends, some correction may be
								// unnecessary
								if (cost <= this.threshold
										&& (this.minimal <= -1 || cost <= this.minimal)) {

									// If the cost is minimal...
									if (cost < minimal_cost)
										minimal_cost = cost;

									// Creating the address
									// address = new Address(String.valueOf(y -
									// 1));
									// Modify the operation sequence
									// r.addAddressPrefix(address);
									temp_al_os = new HashSet<OperationSequence>();
									Iterator<OperationSequence> cellos_iterator = thecells
											.get(i).getSequences().iterator();

									while (cellos_iterator.hasNext()) {
										OperationSequence cellos = cellos_iterator
												.next();

										Iterator<OperationSequence> ros_iterator = r
												.getSequences().iterator();
										while (ros_iterator.hasNext()) {
											OperationSequence ros = ros_iterator
													.next();

											address = new Address(
													String.valueOf(y
															- 1
															+ cellos.getAddressOffset(new Address(
																	String.valueOf(y - 1)))));
											ros.addAddressPrefix(address);

											OperationSequence cellos2 = new OperationSequence(
													cellos);

											cellos2.addOperationSequence(ros);
											temp_al_os.add(cellos2);

										}
									}

									if (minimal_os.containsKey(cost))
										minimal_os.get(cost).addAll(temp_al_os);
									else
										minimal_os.put(cost, temp_al_os);

								}
								/*
								 * // It's a no correction, nothing to do in
								 * fact ! else {
								 * 
								 * // copy the OperationSequences of the up/left
								 * cell temp_al_os = new
								 * HashSet<OperationSequence
								 * >(cell.getCells().get(0).getSequences());
								 * 
								 * if(minimal_os.containsKey(cost))
								 * minimal_os.get(cost).addAll(temp_al_os); else
								 * minimal_os.put(cost, temp_al_os);
								 * 
								 * }
								 */
							}
						}

					}

				}

			}

		}

		// Now let's delete all operation sequence which cost are too big
		minimal_os = deleteHighOperationSequence(minimal_os, this.tree);

		// Now let's add everything :)

		if (minimal_cost != Integer.MAX_VALUE) {
			if (this.minimal == -2) {
				cell = new Cells(minimal_cost, minimal_os);

				matrix.setValue(x, y, cell);
			} else if (minimal_os.get(minimal_cost) != null) {
				HashSet<OperationSequence> result = new HashSet<OperationSequence>();

				result.addAll(minimal_os.get(minimal_cost));

				cell = new Cells(new Cell(minimal_cost, result));
				matrix.setValue(x, y, cell);
			} else {
				cell = new Cells(Integer.MAX_VALUE);
				matrix.setValue(x, y, cell);
			}
		} else {
			cell = new Cells(minimal_cost);
			matrix.setValue(x, y, cell);
		}

		// Return minimal cost
		// DEBUG : System.out.println("x " + x + " y " + y + " cost " +
		// minimal_cost);
		return minimal_cost;
	}

	/**
	 * This function generates every candidate by correcting a tree with the
	 * given character
	 * 
	 * @param c
	 *            Character to use to correct every sub tree
	 * @return Collection of Result
	 * @throws badAddressException
	 * @throws noNodeException
	 * @throws headerNodeException
	 * @throws onlyCharactersException
	 * @throws noMinimalTreeException
	 * @throws badOperationException
	 * @throws addNodeException
	 * @throws removeNodeException
	 * @throws abstractOperationException
	 */

	private Collection<Result> subTreeCorrectionGeneration(char c, Tree t,
			int threshold) throws badAddressException, noNodeException,
			headerNodeException, onlyCharactersException,
			noMinimalTreeException, badOperationException, addNodeException,
			removeNodeException, abstractOperationException {
		Collection<Result> res = null;
		MPair<Character, Tree> couple = new MPair<Character, Tree>(c, t);

		Hashtable<Integer, Collection<Result>> corrections = sub_trees_corrections
				.get(couple);
		if (corrections != null) {
			Enumeration<Integer> e = corrections.keys();
			while (e.hasMoreElements()) {
				int seuil = e.nextElement();
				if (threshold == seuil) {
					res = new ArrayList<Result>(corrections.get(seuil));
				} else if (threshold < seuil) {
					res = new ArrayList<Result>();
					Iterator<Result> it = corrections.get(seuil).iterator();
					while (it.hasNext()) {
						Result r = new Result(it.next());
						if (r.getCost() <= threshold)
							res.add(r);
					}
				}
			}

		} else {
			XmlCorrectorMatrix correction = new XmlCorrectorMatrix(this.ts, t,
					threshold, c);
			Result res2 = null;
			if (this.minimal == -2)
				res = correction.getCorrections();
			else {
				res2 = new Result(correction.getMinimalCorrections());
				res = new ArrayList<Result>();
				res.add(res2);
			}
			corrections = new Hashtable<Integer, Collection<Result>>();
			if (res != null && res.size() > 0) {
				corrections.put(threshold, res);
				sub_trees_corrections.put(couple, corrections);
			}
		}

		return res;
	}

	private HashMap<Integer, HashSet<OperationSequence>> deleteHighOperationSequence(
			HashMap<Integer, HashSet<OperationSequence>> minimal_os, Tree t)
			throws addNodeException, noNodeException, removeNodeException,
			badAddressException, headerNodeException,
			abstractOperationException {
		ArrayList<Integer> costs = new ArrayList<Integer>();
		ArrayList<Tree> trees = new ArrayList<Tree>();
		HashMap<Integer, HashSet<OperationSequence>> res = new HashMap<Integer, HashSet<OperationSequence>>();

		Iterator<Integer> it = minimal_os.keySet().iterator();
		// Calcul des arbres pour cahque sequence et de son cout minimal
		while (it.hasNext()) {
			int cost = it.next();
			Iterator<OperationSequence> it_os = minimal_os.get(cost).iterator();
			while (it_os.hasNext()) {
				OperationSequence os = (OperationSequence) it_os.next();
				Tree t1 = new Tree(t);
				os.make(t1);
				addToArrayCostTree(costs, trees, t1, cost);
			}

		}

		// Suppression des arbres dont le cout est superieur au cout minimal
		it = minimal_os.keySet().iterator();
		while (it.hasNext()) {
			int cost = it.next();
			Iterator<OperationSequence> it_os = minimal_os.get(cost).iterator();
			HashSet<OperationSequence> resos = new HashSet<OperationSequence>();
			while (it_os.hasNext()) {
				OperationSequence os = (OperationSequence) it_os.next();
				Tree t1 = new Tree(t);
				os.make(t1);
				int index = trees.indexOf(t1);
				if (costs.get(index) == cost)
					resos.add(os);
			}
			if (!resos.isEmpty())
				res.put(cost, resos);

		}

		// minimal_os = new HashMap<Integer, HashSet<OperationSequence>>(res);
		return res;

	}

	private void addToArrayCostTree(ArrayList<Integer> costs,
			ArrayList<Tree> trees, Tree t, int cost) {
		boolean trouve = false;
		for (int i = 0; i < trees.size() && !trouve; i++) {
			if (trees.get(i).equals(t)) {
				trouve = true;
				if (cost < costs.get(i))
					costs.set(i, cost);
			}

		}
		if (!trouve) {
			costs.add(cost);
			trees.add(t);
		}
	}

}
