/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.controller;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.lang.management.ManagementFactory;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import fr.ufrt.xmlcorrector.XmlCorrector;
import fr.ufrt.xmlcorrector.XmlCorrectorMatrix;
import fr.ufrt.xmlcorrector.exceptions.XmlControlerException;
import fr.ufrt.xmlcorrector.exceptions.abstractOperationException;
import fr.ufrt.xmlcorrector.exceptions.addNodeException;
import fr.ufrt.xmlcorrector.exceptions.addSchemaException;
import fr.ufrt.xmlcorrector.exceptions.badAddressException;
import fr.ufrt.xmlcorrector.exceptions.badOperationException;
import fr.ufrt.xmlcorrector.exceptions.badSchemaException;
import fr.ufrt.xmlcorrector.exceptions.generatedSchemaException;
import fr.ufrt.xmlcorrector.exceptions.headerNodeException;
import fr.ufrt.xmlcorrector.exceptions.invalidDistanceResultFoundException;
import fr.ufrt.xmlcorrector.exceptions.invalidResultFoundException;
import fr.ufrt.xmlcorrector.exceptions.noNodeException;
import fr.ufrt.xmlcorrector.exceptions.onlyCharactersException;
import fr.ufrt.xmlcorrector.exceptions.removeNodeException;
import fr.ufrt.xmlcorrector.matrix.Result;
import fr.ufrt.xmlcorrector.matrix.TreeEditionMatrix;
import fr.ufrt.xmlcorrector.observer.Observable;
import fr.ufrt.xmlcorrector.observer.Observer;
import fr.ufrt.xmlcorrector.operations.OperationSequence;
import fr.ufrt.xmlcorrector.operations.edit.AddNodeOperation;
import fr.ufrt.xmlcorrector.operations.edit.RelabelNodeOperation;
import fr.ufrt.xmlcorrector.operations.edit.RemoveNodeOperation;
import fr.ufrt.xmlcorrector.schema.TreeSchema;
import fr.ufrt.xmlcorrector.tree.Tree;
import fr.ufrt.xmlcorrector.vocabtranslator.VocabTranslator;

/**
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class XmlCorrectorController implements Observable {

	private ArrayList<Observer> listObserver = new ArrayList<Observer>();

	private final int MAX_TH = 100;

	public TreeSchema ts;
	public Tree t;

	// Param�tres
	public int threshold = -1;
	public String prefix;
	public int type = 1;

	// Options
	public boolean minimal = false;
	public boolean validation_verification = false;
	public boolean distance_verification = false;

	// time
	public long time_start;
	public long time_end;
	public long time_first;
	public long time_cpu;

	// Stats
	public int nb_result = 0;
	public int nb_trees = 0;
	public int nb_correction_sequences = 0;
	public int nb_update_sequences = 0;
	public int nb_edit_sequences = 0;

	// results
	public HashMap<Tree, Hashtable<Integer, Collection<OperationSequence>>> tree_results = new HashMap<Tree, Hashtable<Integer, Collection<OperationSequence>>>();
	public HashMap<Tree, Integer> tree_ids = new HashMap<Tree, Integer>();

	public VocabTranslator tr;

	public XmlCorrectorController() {
		listObserver = new ArrayList<Observer>();

		ts = null;
		t = null;

		// Param�tres
		threshold = -1;
		// prefix;
		type = 1;

		// Options
		minimal = false;
		validation_verification = false;
		distance_verification = false;

		// Stats
		nb_result = 0;
		nb_trees = 0;
		nb_correction_sequences = 0;
		nb_update_sequences = 0;
		nb_edit_sequences = 0;

		// results
		tree_results = new HashMap<Tree, Hashtable<Integer, Collection<OperationSequence>>>();
		tree_ids = new HashMap<Tree, Integer>();
	}

	/**
	 * Generates a TreeSchema using the Element e given. It should have the
	 * correct informations
	 * 
	 * @param e
	 *            Element to use
	 * @throws XmlControlerException
	 * @throws addSchemaException
	 * @throws generatedSchemaException
	 * @throws badSchemaException
	 */
	@SuppressWarnings("unchecked")
	public void parseTreeSchema(Element e) throws XmlControlerException,
			addSchemaException, generatedSchemaException, badSchemaException {

		if (ts != null || e.getChild("root") == null)
			throw new XmlControlerException();

		// Let's take the root so

		String root = e.getChildText("root");

		if (root.length() != 1)
			throw new XmlControlerException();

		ts = new TreeSchema(root.charAt(0));

		// Let's take the rules

		Element c;
		String node;
		String rule;
		Iterator<Element> i = e.getChildren("rule").iterator();

		while (i.hasNext()) {

			c = i.next();

			node = c.getAttributeValue("node");
			if (node == null || node.length() != 1)
				throw new XmlControlerException();

			rule = c.getText();

			ts.addSchema(node.charAt(0), rule);

		}

	}

	/**
	 * Parse the file given, it should be a XmlCorrectorControler conform to the
	 * DTD
	 * 
	 * @param file
	 *            File to use
	 * @throws Exception
	 */
	@SuppressWarnings("unchecked")
	public void parseFile(File file) throws Exception {

		SAXBuilder sxb = new SAXBuilder();

		org.jdom.Document document;
		document = sxb.build(file);

		Element root = document.getRootElement();

		// Si il y a bien une racine
		if (root != null && root.getName().equals("XmlCorrector")) {

			Iterator<Element> i = root.getChildren().iterator();
			Element e;

			while (i.hasNext()) {

				e = i.next();

				if (e.getName().equals("TreeSchemaFile")) {
					String filename = e.getText();

					if (filename != null) {
						tr = new VocabTranslator();
						tr.DTDtoXML(filename, filename + ".xml");

						document = sxb.build(filename + ".xml");

						e = document.getRootElement();

						parseTreeSchema(e);

					}

				} else if (e.getName().equals("TreeFile")) {

					if (t != null)
						throw new XmlControlerException();

					String filename = e.getText();

					if (filename != null)
						t = new Tree(new File(filename), tr);

				} else if (e.getName().equals("Minimal")) {

					if (minimal == false)
						minimal = true;
					else
						throw new XmlControlerException();

				} else if (e.getName().equals("DistanceVerification")) {

					if (distance_verification == false)
						distance_verification = true;
					else
						throw new XmlControlerException();

				} else if (e.getName().equals("ValidationVerification")) {

					if (validation_verification == false)
						validation_verification = true;
					else
						throw new XmlControlerException();

				} else if (e.getName().equals("Minimal")) {

					if (minimal == false)
						minimal = true;
					else
						throw new XmlControlerException();

				} else if (e.getName().equals("ResultFileName")) {

					prefix = e.getText();

				} else if (e.getName().equals("MatrixType"))
					type = 1;
				else if (e.getName().equals("ExplorationType"))
					type = 2;
				else if (e.getName().equals("Threshold")) {

					if (threshold == -1) {

						threshold = Integer.valueOf(e.getText());

						if (threshold > MAX_TH || threshold < -1)
							throw new XmlControlerException();

					} else
						throw new XmlControlerException();

				} else if (e.getName().equals("AddNodeCost")
						|| e.getName().equals("RelabelNodeCost")
						|| e.getName().equals("RemoveNodeCost")) {

					int cost = Integer.valueOf(e.getText());

					if (cost < 0)
						throw new XmlControlerException();
					else if (e.getName().equals("AddNodeCost")) {

						AddNodeOperation.changeOperationCost(cost);

					} else if (e.getName().equals("RelabelNodeCost")) {

						RelabelNodeOperation.changeOperationCost(cost);

					} else if (e.getName().equals("RemoveNodeCost")) {

						RemoveNodeOperation.changeOperationCost(cost);

					}

				}

			}

		}

		else
			throw new Exception();

	}

	/**
	 * Returns an unique id for a Tree. Every tree given is saved and will have
	 * the same id.
	 * 
	 * @param t
	 *            Tree to use to get the ID
	 * @return id of the tree given
	 */
	public Integer getTreeId(Tree t) {

		if (tree_ids.containsKey(t))
			return tree_ids.get(t);
		else {

			tree_ids.put(t, nb_trees);
			nb_trees++;
			return nb_trees - 1;

		}

	}

	/**
	 * Add the tree and the operationSequence that has given it to a Collection
	 * 
	 * @param t
	 *            Tree to Add
	 * @param os
	 *            OperationSequence to do
	 * @param cost
	 *            Cost of the OperationSequence
	 */
	public void addTreeResult(Tree t, OperationSequence os, int cost) {

		// We have some operation sequences for this tree ?

		if (!tree_results.containsKey(t)) {

			tree_results.put(t,
					new Hashtable<Integer, Collection<OperationSequence>>());

		}

		// We have some operation sequence with the same cost ?

		if (!tree_results.get(t).containsKey(cost)) {

			tree_results.get(t).put(cost, new HashSet<OperationSequence>());

		}

		// Ok we can put it now

		tree_results.get(t).get(cost).add(os);

	}

	/**
	 * Verifies that the tree made from the OperationSequence is correct / has
	 * the correct cost
	 * 
	 * @param os
	 *            OperationSequence to use
	 * @param cost
	 *            Cost of the OperationSequence (logically)
	 * @throws invalidResultFoundException
	 *             Throws this if the Tree found isn't valid
	 * @throws invalidDistanceResultFoundException
	 *             Throws this if the distance between the Tree found and the
	 *             original tree isn't the logical cost
	 * @throws addNodeException
	 * @throws noNodeException
	 * @throws removeNodeException
	 * @throws badAddressException
	 * @throws headerNodeException
	 * @throws abstractOperationException
	 * @throws onlyCharactersException
	 * @throws badOperationException
	 */
	public void resultVerification(OperationSequence os, int cost)
			throws invalidResultFoundException, addNodeException,
			noNodeException, removeNodeException, badAddressException,
			headerNodeException, abstractOperationException,
			onlyCharactersException, badOperationException,
			invalidDistanceResultFoundException {

		Tree temp = new Tree(t);
		os.make(temp);

		if (validation_verification && !ts.isValid(temp))
			throw new invalidResultFoundException();
		else if (distance_verification) {

			TreeEditionMatrix tem = new TreeEditionMatrix(t, temp, cost);
			Result temm = tem.compare();
			if (temm == null || temm.getCost() != cost)
				throw new invalidResultFoundException();

		}

		addTreeResult(temp, os, cost);

	}

	/**
	 * Creates the global file result
	 * 
	 * @param n_to_cost
	 *            Hashtable that associates a cost to a file number
	 */
	public void createResultFile(Hashtable<Integer, Integer> n_to_cost) {

		String filename = "results/" + prefix + ".xml";

		try {

			// Let's create the root

			Element racine = new Element("ResultSet");

			Element stats = new Element("Statistics");
			racine.addContent(stats);

			Element parameters = new Element("Parameters");
			racine.addContent(parameters);

			Element result_files = new Element("ResultFiles");
			racine.addContent(result_files);

			Element tree_files = new Element("TreeFiles");
			racine.addContent(tree_files);

			Element child;

			// Then add the type

			child = new Element("Type");
			if (type == 2)
				child.addContent("Exploration");
			else
				child.addContent("Matrix");

			parameters.addContent(child);

			// Then add the threshold

			child = new Element("Threshold");
			child.addContent(String.valueOf(threshold));

			parameters.addContent(child);

			// Minimal

			child = new Element("Minimal");
			if (minimal)
				child.addContent("yes");
			else
				child.addContent("no");

			parameters.addContent(child);

			// Some parameters

			child = new Element("AddNodeCost");
			child.addContent(String.valueOf(AddNodeOperation.cost));

			parameters.addContent(child);

			child = new Element("RelabelNodeCost");
			child.addContent(String.valueOf(RelabelNodeOperation.cost));

			parameters.addContent(child);

			child = new Element("RemoveNodeCost");
			child.addContent(String.valueOf(RemoveNodeOperation.cost));

			parameters.addContent(child);

			// Do the time stats

			child = new Element("Time");
			child.addContent(String.valueOf(time_end - time_start) + "ms");

			stats.addContent(child);

			child = new Element("CPU_Time");
			child.addContent(String.valueOf(time_cpu) + "ms");

			stats.addContent(child);

			// The number found

			child = new Element("DifferentTreesFound");
			child.addContent(String.valueOf(nb_trees));

			stats.addContent(child);

			child = new Element("CorrectOperationSequencesFound");
			child.addContent(String.valueOf(nb_correction_sequences));

			stats.addContent(child);

			child = new Element("UpdateOperationSequencesFound");
			child.addContent(String.valueOf(nb_update_sequences));

			stats.addContent(child);

			child = new Element("EditOperationSequencesFound");
			child.addContent(String.valueOf(nb_edit_sequences));

			stats.addContent(child);

			// File names

			Iterator<Integer> iterator = n_to_cost.keySet().iterator();
			while (iterator.hasNext()) {

				int i = iterator.next();

				child = new Element("ResultFile");
				child.addContent(prefix + "_" + n_to_cost.get(i) + ".xml");
				child.setAttribute("cost", String.valueOf(n_to_cost.get(i)));

				result_files.addContent(child);

			}

			// Tree file names

			Iterator<Tree> iterator_tree = tree_results.keySet().iterator();

			while (iterator_tree.hasNext()) {

				Tree tree = iterator_tree.next();

				child = new Element("TreeResultFile");
				child.addContent(prefix + "_tree_" + getTreeId(tree) + ".xml");
				child.setAttribute("ID", String.valueOf(getTreeId(tree)));
				child.setAttribute(
						"cost",
						String.valueOf(tree_results.get(tree).keySet()
								.iterator().next()));

				tree_files.addContent(child);

			}

			// On finalise

			Document document = new Document(racine);

			XMLOutputter sortie = new XMLOutputter(Format.getPrettyFormat());

			sortie.output(document, new FileOutputStream(filename));

		} catch (java.io.IOException e) {

			System.err.println("Error while creating a result file");
			for (int i = 0; i < e.getStackTrace().length; i++)
				System.out.println(e.getStackTrace()[i]);

		}

	}

	/**
	 * Creates the file result for a given tree
	 * 
	 * @param tree
	 */
	public void createResultFile(Tree tree) {

		String filename = "results/" + prefix + "_tree_" + getTreeId(tree)
				+ ".xml";

		try {

			Element child;

			// On cr�� la racine

			Element racine = new Element("TreeResult");

			// On rajoute le Tree

			child = new Element("Tree");
			child.addContent(tree.toFlatString(tr).replace("<", "[")
					.replace(">", "]"));
			// child.addContent(tree.toString(tr).replace("<", "[").replace(">",
			// "]"));

			racine.addContent(child);

			// On va rajouter toutes les s�quences

			Iterator<Integer> iterator = tree_results.get(tree).keySet()
					.iterator();

			while (iterator.hasNext()) {

				Integer cost = iterator.next();

				Iterator<OperationSequence> ios = tree_results.get(tree)
						.get(cost).iterator();

				while (ios.hasNext()) {

					child = new Element("OperationSequence");
					child.setAttribute("cost", String.valueOf(cost));
					child.addContent(ios.next().toXmlElement(tr));

					racine.addContent(child);

				}

			}

			// On finalise

			Document document = new Document(racine);

			XMLOutputter sortie = new XMLOutputter(Format.getPrettyFormat());

			sortie.output(document, new FileOutputStream(filename));

		} catch (java.io.IOException e) {

			System.out.println("Error while creating a result file");
			for (int i = 0; i < e.getStackTrace().length; i++)
				System.out.println(e.getStackTrace()[i]);

		}

	}

	/**
	 * Creates the result file for the given Result
	 * 
	 * @param resultat
	 *            Result to use
	 * @return Vector that contains the cost and the file number
	 * @throws badOperationException
	 * @throws badAddressException
	 * @throws noNodeException
	 * @throws headerNodeException
	 * @throws addNodeException
	 * @throws removeNodeException
	 * @throws abstractOperationException
	 * @throws invalidResultFoundException
	 * @throws onlyCharactersException
	 * @throws invalidDistanceResultFoundException
	 */
	public Vector<Integer> createResultFile(Result resultat)
			throws badOperationException, badAddressException, noNodeException,
			headerNodeException, addNodeException, removeNodeException,
			abstractOperationException, invalidResultFoundException,
			onlyCharactersException, invalidDistanceResultFoundException {

		String filename = "results/" + prefix + "_" + resultat.getCost()
				+ ".xml";
		nb_result++;

		Vector<Integer> result = new Vector<Integer>();

		try {

			Iterator<OperationSequence> i;
			Iterator<OperationSequence> i2;
			OperationSequence os;
			OperationSequence os2;
			Element child;

			// On cr�� la racine

			Element racine = new Element("Result");

			// On s'occupe du co�t

			child = new Element("Cost");
			child.addContent(String.valueOf(resultat.getCost()));

			racine.addContent(child);

			// On s'occupe des Correct Operation Sequences :o

			child = new Element("CorrectOperationSequences");

			i = resultat.getSequences().iterator();

			while (i.hasNext()) {

				os = i.next();

				if (os.isAbstractStrict()) {
					Tree t2 = new Tree(t);
					os.make(t2);
					child.addContent(os.toXmlElement(t2, tr, getTreeId(t2)));

					nb_correction_sequences++;

				}

			}

			racine.addContent(child);

			// Puis des Update Operation Sequence

			child = new Element("UpdateOperationSequences");

			i = resultat.getSequences().iterator();

			while (i.hasNext()) {

				os = i.next();

				if (os.isAbstractStrict()) {

					i2 = os.abstractionStrictToAbstractOperationSequences()
							.iterator();

					while (i2.hasNext()) {

						os2 = i2.next();

						if (os2.isAbstract()) {
							Tree t2 = new Tree(t);
							os2.make(t2);
							child.addContent(os2.toXmlElement(t2, tr, getTreeId(t2)));
							
							nb_update_sequences++;

						}

					}

				} else if (os.isAbstract()) {
					Tree t2 = new Tree(t);
					os.make(t2);
					child.addContent(os.toXmlElement(t2, tr, getTreeId(t2)));
					
					nb_update_sequences++;

				}

			}

			racine.addContent(child);

			// Puis des Edit Operation Sequence

			child = new Element("EditOperationSequences");

			i = resultat.getSequences().iterator();

			while (i.hasNext()) {

				os = i.next();

				if (os.isAbstract()) {

					i2 = os.abstractionToOperationSequences(new Tree(t))
							.iterator();

					while (i2.hasNext()) {

						os2 = i2.next();
						
						Tree t2 = new Tree(t);
						os2.make(t2);
						child.addContent(os2.toXmlElement(t2, tr, getTreeId(t2)));
						
						nb_edit_sequences++;

						if (validation_verification || distance_verification)
							resultVerification(os2, resultat.getCost());

					}

				} else {
					Tree t2 = new Tree(t);
					os.make(t2);
					child.addContent(os.toXmlElement(t2, tr, getTreeId(t2)));
					
					nb_edit_sequences++;

				}

			}

			racine.addContent(child);

			// On finalise

			Document document = new Document(racine);

			XMLOutputter sortie = new XMLOutputter(Format.getPrettyFormat());

			sortie.output(document, new FileOutputStream(filename));

			// On renvoie l'association

			result.add(nb_result - 1);
			result.add(resultat.getCost());

		} catch (java.io.IOException e) {

			System.out.println("Error while creating a result file");
			for (int i = 0; i < e.getStackTrace().length; i++)
				System.out.println(e.getStackTrace()[i]);

		}

		return result;

	}

	/**
	 * Initialize the variables
	 * @param TreeSchema
	 * @param Tree
	 * @param th
	 * @param mminimal
	 * @param filename
	 * @param addcost
	 * @param relabelcost
	 * @param removecost
	 * @return
	 * @throws Exception
	 */
	public String initVariables(String TreeSchema, String Tree, int th,
			boolean mminimal, String filename, int addcost, int relabelcost,
			int removecost) throws Exception {
		String msg_error = "";
		org.jdom.Document document;
		Element e;
		SAXBuilder sxb = new SAXBuilder();

		/*
		 * if(TreeSchema != null) { document = sxb.build(TreeSchema); e =
		 * document.getRootElement(); parseTreeSchema(e);
		 * 
		 * }
		 */

		if (TreeSchema != null) {
			tr = new VocabTranslator();
			tr.DTDtoXML(TreeSchema, TreeSchema + ".xml");

			document = sxb.build(TreeSchema + ".xml");

			e = document.getRootElement();

			parseTreeSchema(e);
			//We delete the .xml file generate by the DTDtoXML function
			File fTreeSchemaXML = new File(TreeSchema + ".xml");
			if(fTreeSchemaXML.exists()) fTreeSchemaXML.delete();

		}

		if (t != null)
			throw new XmlControlerException();
		// if(Tree != null) t = new Tree(new File(Tree));
		if (Tree != null)
			t = new Tree(new File(Tree), tr);

		if (minimal == false)
			minimal = mminimal;
		else
			throw new XmlControlerException();

		prefix = filename;

		type = 1;

		if (threshold == -1) {

			threshold = th;

			if (threshold > MAX_TH || threshold < -1)
				throw new XmlControlerException();

		} else
			throw new XmlControlerException();

		if (addcost < 0)
			throw new XmlControlerException();
		else
			AddNodeOperation.changeOperationCost(addcost);

		if (relabelcost < 0)
			throw new XmlControlerException();
		else
			RelabelNodeOperation.changeOperationCost(relabelcost);

		if (removecost < 0)
			throw new XmlControlerException();
		else
			RemoveNodeOperation.changeOperationCost(removecost);

		return msg_error;
	}

	/**
	 * Correct the tree with these parameters
	 * @param TreeSchema
	 * @param Tree
	 * @param th
	 * @param mminimal
	 * @param filename
	 * @param addcost
	 * @param relabelcost
	 * @param removecost
	 * @return
	 */
	public String correction(String TreeSchema, String Tree, int th,
			boolean mminimal, String filename, int addcost, int relabelcost,
			int removecost) {
		String msg_error = "";
		// Let's assume the file is ok ?
		try {

			// notifyObserver("Initialisation des variables ...");
			notifyObserver("Initialization of variables ...");
			
			//we create directories logs and results if they don't exist
			File f = new File("results/");
			if(!f.exists()) f.mkdir();
			f = new File("logs/");
			if(!f.exists()) f.mkdir();
			
			long time_aux = System.currentTimeMillis();
			FileOutputStream fos = new FileOutputStream("logs/"	+ filename + "_" + time_aux + ".log", true);
			System.setOut(new PrintStream(fos, true));

			FileOutputStream fos_err = new FileOutputStream("logs/"	+ filename + "_" + time_aux + "_err.log", true);
			System.setErr(new PrintStream(fos_err, true));

			initVariables(TreeSchema, Tree, th, mminimal, filename, addcost, relabelcost, removecost);

			// Let's verify
			if (minimal && (threshold == -1 || threshold == 0)) {
				int cost = ts.getPossibleCostMinimalTree(ts.getRoot());
				// System.out.println("Minimal Tree Cost : " + cost);
				threshold = cost + t.getRemovalCost();

			} else if (threshold == -1 || threshold == 0)
				throw new XmlControlerException();
			if (ts == null || t == null)
				throw new XmlControlerException();
			/*
			 * //notifyObserver("Génération des arbres minimaux ...");
			 * notifyObserver("Generation of minimal trees ..."); // Let's
			 * generate minimal trees long begin_generation =
			 * System.currentTimeMillis();
			 * 
			 * ts.generateMinimalTrees(threshold,minimal);
			 * 
			 * long end_generation = System.currentTimeMillis();
			 * System.out.println(ts);
			 * System.out.println("Génération des arbres minimaux en " +
			 * (end_generation-begin_generation) + " ms");
			 */
			
			

			// Correct !

			XmlCorrector x =  new XmlCorrectorMatrix(ts, t, threshold);

			ArrayList<Result> res;

			System.out.println(System.currentTimeMillis()
					+ "\t Starting generation");

			time_start = System.currentTimeMillis();
			long time_cpu_start = ManagementFactory.getThreadMXBean()
					.getCurrentThreadCpuTime();

			// notifyObserver("Correction de l'arbre ...");
			notifyObserver("Correction of the tree ...");
			// Minimal ?

			if (minimal) {
				boolean trouve = false;
				res = new ArrayList<Result>();
				for (int thd = 1; thd <= threshold && !trouve; thd++) {
					x = new XmlCorrectorMatrix(ts, t, thd);
					res = new ArrayList<Result>();
					res.add(x.getMinimalCorrections());
					if (res.size() > 0) {
						Result rt = res.get(0);
						if (rt.getCost() != -1)
							trouve = true;
					}
				}

			}

			// Not minimal

			else {
				res = new ArrayList<Result>();
				res.addAll(x.getCorrections());

			}

			Iterator<Result> i_r = res.iterator();
			while (i_r.hasNext()) {

				System.out.println(i_r.next());

			}

			time_end = System.currentTimeMillis();
			long time_cpu_end = ManagementFactory.getThreadMXBean().getCurrentThreadCpuTime();
			// Temps en nanoseconde, conversion en ms
			time_cpu = (time_cpu_end - time_cpu_start)	/ ((long) Math.pow(10, 6));
			// time_cpu = time_end - time_start;

			// notifyObserver("Création des résultats ...");
			notifyObserver("Creation of results ...");
			// Creating results

			Iterator<Result> iterator = res.iterator();
			Hashtable<Integer, Integer> n_to_cost = new Hashtable<Integer, Integer>();
			Result resultat;
			while (iterator.hasNext()) {
				resultat = iterator.next();

				Vector<Integer> vector = createResultFile(resultat);
				n_to_cost.put(vector.get(0), vector.get(1));
			}

			// Creating tree results
			Iterator<Tree> tree_iterator = tree_results.keySet().iterator();
			while (tree_iterator.hasNext()) {
				
				Tree tree = tree_iterator.next();

				createResultFile(tree);

			}

			createResultFile(n_to_cost);

			System.out.println(System.currentTimeMillis()
					+ "\t Generation ended");

			fos.close();

		} catch (StackOverflowError e) {

			System.out
					.println("There are too much calculation to do. Stack over flow. Check err log.");
			msg_error += "There are too much calculation to do. Stack over flow. Check err log.";
			for (int i = 0; i < e.getStackTrace().length; i++)
				System.err.println(e.getStackTrace()[i]);

		} catch (Exception e) {

			System.out
					.println("There is an error in TreeSchema or Tree file. Check err log.");
			msg_error += "There is an error in TreeSchema or Tree file. Check err log.";
			for (int i = 0; i < e.getStackTrace().length; i++)
				System.err.println(e.getStackTrace()[i]);

		}
		// System.setOut(System.out);
		return msg_error;

	}

	// **************************************************
	// IMPLÉMENTATION PATTERN OBSERVER
	// **************************************************
	public void addObserver(Observer obs) {
		this.listObserver.add(obs);
	}

	public void notifyObserver(String str) {

		for (Observer obs : listObserver)
			obs.update(str);
	}

	public void removeObserver() {
		listObserver = new ArrayList<Observer>();
	}

}