/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.matrix;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import fr.ufrt.xmlcorrector.exceptions.abstractOperationException;
import fr.ufrt.xmlcorrector.exceptions.addNodeException;
import fr.ufrt.xmlcorrector.exceptions.badAddressException;
import fr.ufrt.xmlcorrector.exceptions.headerNodeException;
import fr.ufrt.xmlcorrector.exceptions.noNodeException;
import fr.ufrt.xmlcorrector.exceptions.removeNodeException;
import fr.ufrt.xmlcorrector.operations.Operation;
import fr.ufrt.xmlcorrector.operations.OperationSequence;
import fr.ufrt.xmlcorrector.tree.Address;
import fr.ufrt.xmlcorrector.tree.Tree;

/**
 * A Cell corresponding to a case of a Matrix
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class Cell {

	Collection<OperationSequence> sequences;
	int cost;

	/**
	 * Constructor that creates a cell by copying the given Cell
	 * 
	 * @param cell  Cell to copy
	 * 
	 */
	public Cell(Cell cell) {

		this.sequences = new ArrayList<OperationSequence>();

		Iterator<OperationSequence> i = cell.sequences.iterator();
		while (i.hasNext())
			addSequence(i.next());

		this.cost = cell.getCost();

	}

	/**
	 * Constructor that creates a cell without any OperationSequence but with a
	 * cost
	 * 
	 * @param cost
	 *            Cost of the cell
	 */
	public Cell(int cost) {

		this.sequences = new ArrayList<OperationSequence>();
		this.cost = cost;

	}

	/**
	 * Constructor that creates a cell with OperationSequences but with a cost
	 * 
	 * @param cost
	 *            Cost of the cell
	 * @param sequences
	 *            ArrayList of OperationSequences
	 */
	public Cell(int cost, ArrayList<OperationSequence> sequences) {

		this.sequences = new ArrayList<OperationSequence>(sequences);
		this.cost = cost;

	}

	/**
	 * Constructor that creates a cell with OperationSequences but with a cost
	 * 
	 * @param cost
	 *            Cost of the cell
	 * @param sequences
	 *            ArrayList of OperationSequences
	 */
	public Cell(int cost, HashSet<OperationSequence> sequences) {

		this.sequences = new ArrayList<OperationSequence>(sequences);
		this.cost = cost;

	}

	/**
	 * Adds an OperationSequence
	 * 
	 * @param sequence
	 *            OperationsSquence to add
	 */
	public void addSequence(OperationSequence sequence) {

		this.sequences.add(new OperationSequence(sequence));

	}

	/**
	 * Adds a collection of OperationSequence to the result
	 * 
	 * @param sequences
	 *            OperationSequences to add
	 */
	public void addSequences(Collection<OperationSequence> sequences) {

		Iterator<OperationSequence> i = sequences.iterator();
		while (i.hasNext())
			addSequence(i.next());

	}

	/**
	 * Returns the cost to make the operation sequences
	 * 
	 * @return unique cost
	 */
	public int getCost() {

		return this.cost;

	}

	/**
	 * Changes the current cost with the cost given
	 * 
	 * @param cost
	 *            Cost to use
	 */
	public void setCost(int cost) {

		this.cost = cost;

	}

	/**
	 * Returns the OperationSequence at position / null if there isn't any
	 * 
	 * @param position
	 *            Position of the OperationSequence to return
	 * @return Wanted Operation Sequence
	 */
	public OperationSequence getSequence(int position) {

		int i = 0;
		Iterator<OperationSequence> iterator = this.sequences.iterator();

		while (iterator.hasNext()) {

			if (i == position)
				return new OperationSequence(iterator.next());
			else {

				iterator.next();
				i++;

			}

		}

		return null;

	}

	/**
	 * Returns the Collection of OperationSequence
	 * 
	 * @return An ArrayList of OperationSequences
	 */
	public Collection<OperationSequence> getSequences() {

		ArrayList<OperationSequence> result = new ArrayList<OperationSequence>();
		Iterator<OperationSequence> i = this.sequences.iterator();

		while (i.hasNext()) {

			result.add(new OperationSequence(i.next()));

		}

		return result;

	}

	/**
	 * Use the given Tree to make the Operations of the first OperationSequence
	 * and returns the modified Tree. It doesn't modify the given Tree.
	 * 
	 * @param t
	 *            Tree to use to make every Operations
	 * @return the Tree after every modifications
	 * @throws addNodeException
	 *             Throws this Exception if there is an error while trying to
	 *             add a Node
	 * @throws noNodeException
	 *             Throws this Exception if an Operation tries to access to a
	 *             Node that doesn't exist
	 * @throws removeNodeException
	 *             Throws this Exception if there is an error while trying to
	 *             remove a Node
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 * @throws headerNodeException
	 *             Throws this exception is the header of the new node is ""
	 * @throws abstractOperationException
	 *             Throws this exception if the is an abstract operation in the
	 *             sequence like AddMinimalTreeOperation
	 */
	public Tree make(Tree t) throws addNodeException, noNodeException,
			removeNodeException, badAddressException, headerNodeException,
			abstractOperationException {

		Tree tree = new Tree(t);

		if (this.sequences.size() > 0)
			this.sequences.iterator().next().make(t);

		return tree;

	}

	/**
	 * Adds the OperationSequence given at the end of every OperationSequence of
	 * the Result. Adds it as OperationSequence if the Result hasn't any
	 * OperationSequence
	 * 
	 * @param os
	 *            OperationSequence to add at the end of every OperationSequence
	 */
	public void addToSequences(OperationSequence os) {

		if (this.sequences.size() > 0) {

			Iterator<OperationSequence> iterator = this.sequences.iterator();

			while (iterator.hasNext()) {

				iterator.next().addOperationSequence(os);

			}

		}

		else
			addSequence(os);

	}

	/**
	 * Adds the Operation given at the end of every OperationSequence of the
	 * Result. Adds it as OperationSequence if the Result hasn't any
	 * OperationSequence
	 * 
	 * @param o Operation to add at the end of every OperationSequence
	 */
	public void addToSequences(Operation o) {

		if (this.sequences.size() > 0) {

			Iterator<OperationSequence> iterator = this.sequences.iterator();

			while (iterator.hasNext()) {

				iterator.next().addOperation(o);

			}

		}

		else {

			OperationSequence os = new OperationSequence();
			os.addOperation(o);

			addSequence(os);

		}

	}

	/**
	 * Changes the Address of every OperationSequence by adding the given
	 * Address at the begin of their Addresses
	 * 
	 * @param reference
	 *            Address to add at the begin of every Operation's Address
	 */
	public void addAddressPrefix(Address reference) {

		Iterator<OperationSequence> i = this.sequences.iterator();
		while (i.hasNext()) {

			i.next().addAddressPrefix(reference);

		}

	}

	public String toString() {

		// return
		// String.valueOf(this.cost)+" ("+String.valueOf(this.sequences.size())+")";
		return String.valueOf(this.cost);

	}

	public Cell clone() {

		return new Cell(this);

	}

	public String toFlatString() {

		return String.valueOf(this.cost) + ": " + this.sequences;
	}

}
