/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.matrix;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

import fr.ufrt.xmlcorrector.exceptions.*;
import fr.ufrt.xmlcorrector.operations.*;
import fr.ufrt.xmlcorrector.operations.edit.AddNodeOperation;
import fr.ufrt.xmlcorrector.operations.edit.RelabelNodeOperation;
import fr.ufrt.xmlcorrector.operations.update.AddTreeOperation;
import fr.ufrt.xmlcorrector.operations.update.RemoveTreeOperation;
import fr.ufrt.xmlcorrector.tree.Address;
import fr.ufrt.xmlcorrector.tree.Tree;

/**
 * A Tree Edition Matrix for measuring the distance between two trees.
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class TreeEditionMatrix {

	Matrix<Cell> matrix;
	Tree tree;
	Tree reference;
	int max_cost;

	/**
	 * Creates a TreeEditionMatrix where tree should be correct with reference.
	 * The correction should be cheaper than cost;
	 * 
	 * @param tree
	 *            Tree to correct
	 * @param reference
	 *            Tree to use as correction
	 * @param max_cost
	 *            Maximal cost of the correction
	 */
	public TreeEditionMatrix(Tree tree, Tree reference, int max_cost) {

		// Size of the matrix
		int width = (reference.isEmpty()) ? 1 : reference.getRoot()
				.getChildrenNumber() + 1;
		int height = (tree.isEmpty()) ? 1
				: tree.getRoot().getChildrenNumber() + 1;

		// Matrix & cost
		this.matrix = new SimpleMatrix<Cell>(width, height);
		this.max_cost = max_cost;

		// Trees
		this.tree = tree;
		this.reference = reference;

	}

	/**
	 * Method used to copy a list without shared memory
	 * 
	 * @param source
	 *            List to copy
	 * @return copy of the array list
	 */
	private ArrayList<OperationSequence> arrayListCopy(
			Collection<OperationSequence> source) {

		ArrayList<OperationSequence> result = new ArrayList<OperationSequence>();
		Iterator<OperationSequence> i = source.iterator();

		while (i.hasNext()) {

			result.add(new OperationSequence(i.next()));

		}

		return result;

	}

	/**
	 * Method used to add every item of c to the collection hs
	 * 
	 * @param hs
	 *            Collection where the items should be added
	 * @param c
	 *            Collection of items that should be added
	 */
	private void hashSetAddAll(Collection<OperationSequence> hs,
			Collection<OperationSequence> c) {

		Iterator<OperationSequence> i = c.iterator();
		OperationSequence os;

		while (i.hasNext()) {

			os = i.next();
			if (!hs.contains(os))
				hs.add(os);

		}

	}

	/**
	 * Method that generates the cell at (x, y), it calculates the cost and the
	 * OperationSequence to get there
	 * 
	 * @param x
	 *            abscissa of the cell
	 * @param y
	 *            ordinate of the cell
	 * @return Cell to add
	 * @throws noNodeException
	 *             Throws this exception if there is an access to an unexistant
	 *             node
	 * @throws badAddressException
	 *             Throws this exception if there is a problem with addresses
	 * @throws headerNodeException
	 *             Throws this exception if there is a problem with node's
	 *             header
	 * @throws badOperationException
	 *             Throws this exception if there is a problem with operation
	 */
	private Cell getCell(int x, int y) throws noNodeException,
			badAddressException, headerNodeException, badOperationException {

		Cell c = null;

		// First, we verify we are not in a "special" Cell

		if ((x == 0 || y == 0) && x != y) {

			int cost = 0;
			OperationSequence os;
			Cell cell;

			// We are at (x, 0)

			if (y == 0) {

				cell = this.matrix.getValue(x - 1, y);
				os = cell.getSequence(0);

				cost = cell.getCost()
						+ reference.getNodeAsTree(
								new Address(String.valueOf(x - 1)))
								.getCreationCost();

				os.addOperation(new AddTreeOperation(new Address(String
						.valueOf(x - 1)), reference.getNodeAsTree(new Address(
						String.valueOf(x - 1)))));

			}

			// We are at (0, y)

			else {

				cell = this.matrix.getValue(x, y - 1);
				os = cell.getSequence(0);

				cost = cell.getCost()
						+ tree.getNodeAsTree(new Address(String.valueOf(y - 1)))
								.getRemovalCost();

				Address a = new Address(
						String.valueOf(y
								- 1
								+ os.getAddressOffset(new Address(String
										.valueOf(y - 1)))));
				os.addOperation(new RemoveTreeOperation(a));

			}

			// Let's create the cell

			c = new Cell(cost);
			c.addToSequences(os);

		}

		// It's okay, we're in a normal cell

		else {

			int cost;
			int iterator;
			int minimal_cost = -1;
			Cell cell;
			HashMap<Integer, HashSet<OperationSequence>> minimal_os = new HashMap<Integer, HashSet<OperationSequence>>();
			HashSet<OperationSequence> temp_al_os;
			Iterator<OperationSequence> hash_iterator;

			Address address;

			// First: Look upside !

			if (y - 1 >= 0) {

				// First get the cell
				cell = this.matrix.getValue(x, y - 1);

				// The address of the Node to add
				address = new Address(String.valueOf(y - 1));

				// Get the total cost...
				cost = cell.getCost()
						+ this.tree.getNodeAsTree(address).getRemovalCost();

				// If the cost si minimal, let's save it
				if (cost < minimal_cost || minimal_cost == -1)
					minimal_cost = cost;

				// if it is minimal, we have to asve the sequence...
				if (cost <= minimal_cost) {

					// let's copy every sequence of the previous cell
					temp_al_os = new HashSet<OperationSequence>(
							cell.getSequences());
					hash_iterator = temp_al_os.iterator();

					// We need a var
					OperationSequence temp_os;

					// For every sequence, let's add our tree...
					while (hash_iterator.hasNext()) {

						temp_os = hash_iterator.next();

						// let's create the address with the offset...
						address = new Address(String.valueOf(y
								- 1
								+ temp_os.getAddressOffset(new Address(String
										.valueOf(y - 1)))));

						// and add it to the operations :/
						temp_os.addOperation(new RemoveTreeOperation(address));

					}

					if (minimal_os.containsKey(cost))
						minimal_os.get(cost).addAll(temp_al_os);
					else
						minimal_os.put(cost, temp_al_os);

				}

			}

			// Second: Look leftside !

			if (x - 1 >= 0) {

				// first get the cell
				cell = this.matrix.getValue(x - 1, y);

				// then the address of the node to remove
				address = new Address(String.valueOf(x - 1));

				// the cost
				cost = cell.getCost()
						+ this.reference.getNodeAsTree(address)
								.getCreationCost();

				// if the cost is minimal...
				if (cost < minimal_cost || minimal_cost == -1)
					minimal_cost = cost;

				// then we should save results
				if (cost <= minimal_cost) {

					// first copy the OperationSequences of the left cell
					temp_al_os = new HashSet<OperationSequence>(
							cell.getSequences());
					hash_iterator = temp_al_os.iterator();

					// We need a var
					OperationSequence temp_os;

					// On every of the cell, add the Operation
					while (hash_iterator.hasNext()) {

						temp_os = hash_iterator.next();

						// let's add the operation
						temp_os.addOperation(new AddTreeOperation(address,
								this.reference.getNodeAsTree(new Address(String
										.valueOf(x - 1)))));

					}

					if (minimal_os.containsKey(cost))
						minimal_os.get(cost).addAll(temp_al_os);
					else
						minimal_os.put(cost, temp_al_os);

				}

			}

			// Third, up-left side !

			if (y - 1 >= 0 && x - 1 >= 0) {

				// First we get the Cell
				cell = this.matrix.getValue(x - 1, y - 1);

				// Then we create a recursive call comparing the two trees
				TreeEditionMatrix tem = new TreeEditionMatrix(
						tree.getNodeAsTree(new Address(String.valueOf(y - 1))),
						reference.getNodeAsTree(new Address(String
								.valueOf(x - 1))),
						(this.max_cost - cell.getCost()));
				Result temp = tem.compare();

				// If the two trees can be compared, let's add the results ?
				if (temp != null) {

					// Geting the costs
					cost = temp.getCost() + cell.getCost();

					// If the cost is minimal...
					if (cost < minimal_cost || minimal_cost == -1)
						minimal_cost = cost;

					// If it is minimal, we have to get the
					if (cost <= minimal_cost) {

						// Creating the address
						address = new Address(String.valueOf(y - 1));

						// Okay we will have to use multiple Lists... It's very
						// hard to get them clearly
						temp_al_os = (HashSet<OperationSequence>) new HashSet<OperationSequence>(); // temp_al_os
																									// will
																									// be
																									// the
																									// finial
																									// list
																									// that
																									// will
																									// be
																									// adden
																									// to
																									// the
																									// results..
						ArrayList<OperationSequence> temp_al_os_2; // temp_al_os_2
																	// are the
																	// OperationSequence
																	// that were
																	// in the
																	// previous
																	// Cell
						ArrayList<OperationSequence> temp_al_os_3; // tmep_al_os_3
																	// are the
																	// OperationSequence
																	// that came
																	// from our
																	// recursive
																	// call...

						// We will have multiple Prefix to add... so let's get
						// some var
						OperationSequence temp_os;

						// First, let's get the OperationSequence of the
						// recursive call..
						temp_al_os_3 = (ArrayList<OperationSequence>) temp
								.getSequences();

						// Then for every OperationSequence of the call...
						for (int iterator_2 = 0; iterator_2 < temp_al_os_3
								.size(); iterator_2++) {

							// Lets make a first copy of the OperationSequences
							// of the cell at x-1; y-1
							temp_al_os_2 = arrayListCopy(cell.getSequences());

							// ok so for every OperationSequence of the cell
							// x-1/y-1...
							for (iterator = 0; iterator < temp_al_os_2.size(); iterator++) {

								// let's copy the OperationSequence to add
								temp_os = new OperationSequence(
										temp_al_os_3.get(iterator_2));

								// Let's add some offset :(
								address = new Address(
										String.valueOf(y
												- 1
												+ temp_al_os_2
														.get(iterator)
														.getAddressOffset(
																new Address(
																		String.valueOf(y - 1)))));
								temp_os.addAddressPrefix(address);

								// and let's add it to the operationsequence of
								// the cell x-1/y-1
								temp_al_os_2.get(iterator)
										.addOperationSequence(temp_os);

							}

							// Now we add everything to the results
							hashSetAddAll(temp_al_os, temp_al_os_2);

						}

						if (minimal_os.containsKey(cost))
							hashSetAddAll(minimal_os.get(cost), temp_al_os);
						else
							minimal_os.put(cost, temp_al_os);

					}

				}

			}

			// Now let's add everything :)

			HashSet<OperationSequence> result = new HashSet<OperationSequence>();
			result.addAll(minimal_os.get(minimal_cost));

			c = new Cell(minimal_cost, result);

		}

		return c;

	}

	/**
	 * Method used to generate every cell that has x as abscissa
	 * 
	 * @param x
	 *            Abscissa of the cells
	 */

	private void compareRecursive(int x) throws noNodeException,
			badAddressException, headerNodeException, badOperationException {

		Cell c;
		int min_cost = (x > 0) ? -1 : this.matrix.getValue(0, 0).getCost();
		int start = (x > 0) ? 0 : 1;

		// Do the cells

		for (int j = start; j < this.matrix.getHeight(); j++) {

			c = getCell(x, j);

			this.matrix.setValue(x, j, c);

			if (min_cost == -1 || min_cost > c.getCost())
				min_cost = c.getCost();

		}

		// Recursive ?

		if (min_cost <= this.max_cost && x < this.matrix.getWidth() - 1)
			compareRecursive(x + 1);

	}

	/**
	 * Method that will calculates every cell of the matrix and return the
	 * result. null if there is no correction
	 * 
	 * @return the result cell, or null if there is no correction
	 * @throws headerNodeException
	 * @throws badAddressException
	 * @throws noNodeException
	 * @throws badOperationException
	 */

	public Result compare() throws headerNodeException, badAddressException,
			noNodeException, badOperationException {

		// We calculate for the root

		Cell c = null;

		if (this.tree.isEmpty() && this.reference.isEmpty()) {

			c = new Cell(0);

		}

		else if (this.tree.isEmpty() || this.reference.isEmpty()) {

			OperationSequence os = new OperationSequence();

			if (this.tree.isEmpty()) {

				os.addOperation(new AddNodeOperation(new Address(""),
						this.reference.getRoot().getHeader()));

				c = new Cell(AddNodeOperation.cost);
				c.addSequence(os);

			} else {

				Operation o = new RemoveTreeOperation(new Address(""));

				os.addOperation(o);

				c = new Cell(o.getCost(this.tree));
				c.addSequence(os);

				for (int i = 0; i < this.matrix.getHeight(); i++)
					this.matrix.setValue(0, i, new Cell(c));

				return new Result(c);

			}

		}

		else if (this.tree.getRoot().getHeader() != this.reference.getRoot()
				.getHeader()) {

			OperationSequence os = new OperationSequence();
			os.addOperation(new RelabelNodeOperation(new Address(""),
					this.reference.getRoot().getHeader()));

			c = new Cell(RelabelNodeOperation.cost);
			c.addSequence(os);

		}

		else {

			c = new Cell(0);

			OperationSequence os = new OperationSequence();
			c.addSequence(os);

		}

		// We add the first cell

		this.matrix.setValue(0, 0, c);

		// We launch the recusivity

		compareRecursive(0);

		return (this.matrix.getValue((this.matrix.getWidth() - 1),
				(this.matrix.getHeight() - 1)) == null || this.matrix.getValue(
				(this.matrix.getWidth() - 1), (this.matrix.getHeight() - 1))
				.getCost() > this.max_cost) ? null : new Result(
				this.matrix.getValue((this.matrix.getWidth() - 1),
						(this.matrix.getHeight() - 1)));

	}

	/**
	 * Returns a copy of the matrix
	 * 
	 * @return a Matrix<Cell> copied from the one of the comparaison
	 */
	public Matrix<Cell> getMatrix() {

		return new SimpleMatrix<Cell>(this.matrix);

	}

	public String toString() {

		String resultat = "---------- Matrice -----------\n"
				+ this.matrix.toString() + "\n";
		resultat = resultat + "---------- Solution -----------\n";

		if (this.matrix.getValue((this.matrix.getWidth() - 1),
				(this.matrix.getHeight() - 1)) == null)
			resultat = resultat + "aucune solution";
		else {

			Iterator<OperationSequence> i = this.matrix
					.getValue((this.matrix.getWidth() - 1),
							(this.matrix.getHeight() - 1)).getSequences()
					.iterator();
			while (i.hasNext()) {

				resultat = resultat + i.next() + "\n\n";

			}

		}

		return resultat;

	}

}
