package fr.ufrt.xmlcorrector.operations;

import org.jdom.Element;

import fr.ufrt.xmlcorrector.exceptions.abstractOperationException;
import fr.ufrt.xmlcorrector.exceptions.addNodeException;
import fr.ufrt.xmlcorrector.exceptions.badAddressException;
import fr.ufrt.xmlcorrector.exceptions.badOperationException;
import fr.ufrt.xmlcorrector.exceptions.headerNodeException;
import fr.ufrt.xmlcorrector.exceptions.noNodeException;
import fr.ufrt.xmlcorrector.exceptions.removeNodeException;
import fr.ufrt.xmlcorrector.tree.Address;
import fr.ufrt.xmlcorrector.tree.Tree;
import fr.ufrt.xmlcorrector.vocabtranslator.VocabTranslator;

public abstract class Operation {

	protected OperationType operation;
	protected Address address;

	public enum OperationType {
		relabelNode, addNode, removeNode, removeTree, replaceTree, addTree, addMinimalTree, CorrectTree
	}

	/**
	 * Gives the label of the Operation
	 * 
	 * @return a String that can be relabelNode, addNode, removeNode,
	 *         removeTree, replaceTree or addTree
	 */
	public String getOperationString() {

		return this.operation.toString();

	}

	/**
	 * Gives the label of the Operation
	 * 
	 * @return an OperationType
	 */
	public OperationType getOperationType() {

		return this.operation;

	}

	/**
	 * Returns a copy of the Address where the Operation should be made
	 * 
	 * @return Address of the Operation
	 */
	public Address getAddress() {

		return new Address(this.address);

	}

	/**
	 * Changes the Address of the Operation with the one given
	 * 
	 * @param address
	 *            Address to use
	 */
	public void setAddress(Address address) {

		this.address = new Address(address);

	}

	/**
	 * Puts the prefix given at the begin of the Address of the Operation
	 * 
	 * @param prefix
	 *            new begin of the Address of the Operation
	 */
	public void addAddressPrefix(Address prefix) {

		this.address.addPrefix(prefix);

	}

	/**
	 * Calculates the cost of the Operation on the given Tree
	 * 
	 * @param tree
	 *            Tree where the Operation should be made
	 * @return Cost of the Operation (integer)
	 * @throws badOperationException
	 *             Throws this Exception if the Operation isn't ok
	 * @throws noNodeException
	 *             Throws this exception if the Operation contains an Address
	 *             that doesn't exist in the given Tree
	 */
	public abstract int getCost(Tree tree) throws badOperationException,
			noNodeException;

	/**
	 * Makes the Operation on the given Tree
	 * 
	 * @param tree
	 *            Tree where the Operation should be done
	 * @throws addNodeException
	 *             Throws this Exception if the Operation implicates the
	 *             creation of a Node that can't be done
	 * @throws noNodeException
	 *             Throws this Exception if the Operation implicates an access
	 *             to a Node that doesn't exist
	 * @throws removeNodeException
	 *             Throws this Exception if the Operation tries to remove a Node
	 *             that isn't a Leaf
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 * @throws headerNodeException
	 *             Throws this exception is the header of the new node is ""
	 * @throws abstractOperationException
	 *             Throws this exception if the Operation is an Abstract
	 *             operation like AddMinimalTreeOperation
	 */
	public abstract void make(Tree tree) throws addNodeException,
			noNodeException, removeNodeException, badAddressException,
			headerNodeException, abstractOperationException;

	/**
	 * Converts the operation to a string like OperationName[attributes]
	 */
	public abstract String toString();

	/**
	 * Converts the operation to a string like OperationName[attributes]
	 */
	public String toCompleteString() {

		return this.toString();

	}

	/**
	 * Creates a copy of the Operation
	 */
	public abstract Operation clone();

	/**
	 * Tests if the Operation is an UpdateOperation or an EditOperation.
	 * 
	 * @return True if the Operation is Abstract (UpdateOperation)
	 */
	public abstract boolean isAbstract();

	/**
	 * Tests if the Operation is an Abstract Strict Operation, meaning a
	 * compilation of multiple UpdateOperation that have the same cost but not
	 * the same result
	 * 
	 * @return True if the Operation is an Abstract Strict Operation, else false
	 */
	public abstract boolean isAbstractStrict();

	/**
	 * Tests if the object given is from the same type and has the same
	 * hashCode, this means they have the same values for attributes
	 */
	@Override
	public abstract boolean equals(Object o);

	/**
	 * Returns an unique hashCode depending on the values of the attributes
	 */
	@Override
	public abstract int hashCode();

	public abstract Element toXmlElement();

	public abstract Element toXmlElement(VocabTranslator vt);

}
