package fr.ufrt.xmlcorrector.operations;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import org.jdom.Element;

import fr.ufrt.xmlcorrector.exceptions.abstractOperationException;
import fr.ufrt.xmlcorrector.exceptions.addNodeException;
import fr.ufrt.xmlcorrector.exceptions.badAddressException;
import fr.ufrt.xmlcorrector.exceptions.badOperationException;
import fr.ufrt.xmlcorrector.exceptions.headerNodeException;
import fr.ufrt.xmlcorrector.exceptions.noNodeException;
import fr.ufrt.xmlcorrector.exceptions.removeNodeException;
import fr.ufrt.xmlcorrector.operations.abstractupdate.AbstractUpdateOperation;
import fr.ufrt.xmlcorrector.operations.update.UpdateOperation;
import fr.ufrt.xmlcorrector.tree.Address;
import fr.ufrt.xmlcorrector.tree.Tree;
import fr.ufrt.xmlcorrector.vocabtranslator.VocabTranslator;

public class OperationSequence {

	ArrayList<Operation> operations;

	/**
	 * Constructor that copies a given sequence in the new object
	 * 
	 * @param sequence
	 *            OperationSequence to copy
	 */
	public OperationSequence(OperationSequence sequence) {

		this.operations = new ArrayList<Operation>();

		Iterator<Operation> i = sequence.operations.iterator();
		while (i.hasNext())
			addOperation(i.next());

	}

	/**
	 * Constructor that creates an empty sequence
	 */
	public OperationSequence() {

		this.operations = new ArrayList<Operation>();

	}

	/**
	 * Adds an Operation at the end of the sequence
	 * 
	 * @param operation
	 *            Operation to add
	 */
	public void addOperation(Operation operation) {

		this.operations.add(operation.clone());

	}

	/**
	 * Adds an Operation at the given position in the sequence
	 * 
	 * @param operation
	 *            Operation to add
	 * @param position
	 *            Position where the Operation should be added
	 */
	public void addOperation(Operation operation, int position) {

		this.operations.add(position, operation.clone());

	}

	/**
	 * Adds every a copy Operation of the OperatiosSequence at the end of the
	 * current sequence
	 * 
	 * @param sequence
	 *            OperationSequence to use
	 */
	public void addOperationSequence(OperationSequence sequence) {

		Iterator<Operation> i = sequence.operations.iterator();
		while (i.hasNext())
			addOperation(i.next().clone());

	}

	/**
	 * Adds a copy of every Operation of the OperatiosSequence at the end of the
	 * current sequence
	 * 
	 * @param sequence
	 *            OperationSequence to use
	 * @param position
	 *            Where the operations should be added
	 */
	public void addOperationSequence(OperationSequence sequence, int position) {

		int count = 0;
		Iterator<Operation> i = sequence.operations.iterator();
		while (i.hasNext())
			addOperation(i.next().clone(), (position + (count++)));

	}

	/**
	 * Calculates the global cost of the sequence for doing every Operations on
	 * the Tree given
	 * 
	 * @param t
	 *            Tree that should be used to do the sequence
	 * @return The global cost of the sequence
	 * @throws badOperationException
	 *             Throws this Exception if an error occurs while doing the
	 *             Operations
	 * @throws noNodeException
	 *             Throws t his Exception if an Operation implicates the use of
	 *             a Node that doesn't exists
	 */
	public int getCost(Tree t) throws badOperationException, noNodeException {

		int cost = 0;

		Iterator<Operation> i = this.operations.iterator();
		while (i.hasNext()) {

			cost = cost + i.next().getCost(t);

		}

		return cost;

	}

	/**
	 * Makes every Operation of the OperationSequence on the Tree given
	 * 
	 * @param t
	 *            Tree where the Operations should be done
	 * @throws addNodeException
	 *             Throws this Exception if there is an error while trying to
	 *             add a Node
	 * @throws noNodeException
	 *             Throws this Exception if an Operation tries to access to a
	 *             Node that doesn't exist
	 * @throws removeNodeException
	 *             Throws this Exception if there is an error while trying to
	 *             remove a Node
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 * @throws headerNodeException
	 *             Throws this exception is the header of the new node is ""
	 * @throws abstractOperationException
	 *             Throws this exception if the is an abstract operation in the
	 *             sequence like AddMinimalTreeOperation
	 */
	public void make(Tree t) throws addNodeException, noNodeException,
			removeNodeException, badAddressException, headerNodeException,
			abstractOperationException {

		Iterator<Operation> i = this.operations.iterator();
		while (i.hasNext()) {

			i.next().make(t);

		}

	}

	/**
	 * Changes the Address of every Operations of the Sequence by adding the
	 * given Address at the begin of their Addresses
	 * 
	 * @param reference
	 *            Address to add at the begin of every Operation's Address
	 */
	public void addAddressPrefix(Address reference) {

		if (!reference.isRoot()) {

			Iterator<Operation> i = this.operations.iterator();
			while (i.hasNext()) {

				i.next().addAddressPrefix(reference);

			}

		}

	}

	/**
	 * Return True if there is an Operation in the Sequence that isn't an
	 * EditOperation
	 * 
	 * @return True if the OperationSequence is someway abstract
	 */
	public boolean isAbstract() {

		Iterator<Operation> i = this.operations.iterator();
		while (i.hasNext()) {

			if (i.next().isAbstract())
				return true;

		}

		return false;

	}

	/**
	 * Return True if there is an Operation in the Sequence that isn't an
	 * Abstract UpdateOperation
	 * 
	 * @return True if the OperationSequence is someway abstract strict
	 */
	public boolean isAbstractStrict() {

		Iterator<Operation> i = this.operations.iterator();
		while (i.hasNext()) {

			if (i.next().isAbstractStrict())
				return true;

		}

		return false;

	}

	/**
	 * Returns the number of operation contained by the sequence
	 * 
	 * @return number of operations
	 */
	public int getOperationNumber() {

		return this.operations.size();

	}

	/**
	 * Returns the Operation that is at the position given
	 * 
	 * @param i
	 *            Position, from 0 to number-1
	 * @return Operation
	 */
	public Operation getOperation(int i) {

		return this.operations.get(i).clone();

	}

	/**
	 * Removes the Operation that is at the given position
	 * 
	 * @param i
	 *            Position between 0 and total-1
	 */
	public void removeOperation(int i) {

		this.operations.remove(i);

	}

	/**
	 * Calculates the offset generated by the making of the operation already
	 * done when the next one to be added will be make. It depends on the level
	 * of the address given
	 * 
	 * @param address
	 *            Address where the next operation to be add should interfer
	 * @return offset to add to this address to make it valid
	 */
	public int getAddressOffset(Address address) {

		int offset = 0;

		// Our sequence is empty
		if (this.getOperationNumber() == 0)
			return offset;

		// Let check every Operation
		Iterator<Operation> i = this.operations.iterator();
		Operation o;
		Operation.OperationType ot;

		// For every operation
		while (i.hasNext()) {

			o = i.next();
			ot = o.getOperationType();

			// if it's an add, +1 !
			if (ot == Operation.OperationType.addMinimalTree
					|| ot == Operation.OperationType.addNode
					|| ot == Operation.OperationType.addTree) {

				if (address.isSameLevel(o.getAddress()))
					offset++;

			}

			// if it's a remove, -1 !
			else if (ot == Operation.OperationType.removeNode
					|| ot == Operation.OperationType.removeTree) {

				if (address.isSameLevel(o.getAddress()))
					offset--;

			}

		}

		return offset;

	}

	/**
	 * Returns a String that is like this: sequence[list of operations]
	 */
	public String toString() {

		String s = "sequence[";

		Iterator<Operation> i = this.operations.iterator();
		while (i.hasNext())
			s = s + i.next();

		s = s + "]";

		return s;

	}

	/**
	 * Returns true if the given object is an OperationSequence with the same
	 * hashCode
	 */
	@Override
	public boolean equals(Object o) {

		if (o instanceof OperationSequence)
			return o.hashCode() == this.hashCode();
		else
			return false;

	}

	/**
	 * Returns an unique hashCode
	 */
	@Override
	public int hashCode() {

		String result = "";

		Iterator<Operation> i = this.operations.iterator();
		while (i.hasNext())
			result = "o" + result + i.next().hashCode();

		return result.hashCode();

	}

	/**
	 * Method used to generate OperationSequences that isn't more than Abstract.
	 * It unabstract the Operation i from the current OperationSequence and make
	 * the cartesian product with c
	 * 
	 * @param i
	 *            Number of Operation to unabstract
	 * @param c
	 *            Collection that contains OperationSequence unabstracted given
	 *            by the i-x operations (x > 0)
	 * @return Cartesian product
	 */
	private Collection<OperationSequence> recursiveAbstractionStrictToAbstractOperationSequences(
			int i, Collection<OperationSequence> c) {

		// If there isn't any Operation at i
		if (i >= this.operations.size())
			return c;

		// Else, get the operation

		Operation o = this.getOperation(i);

		// if the operation is abstract strict

		if (o instanceof AbstractUpdateOperation) {

			AbstractUpdateOperation o2 = (AbstractUpdateOperation) o;

			Collection<OperationSequence> c2 = o2
					.abstractionStrictToAbstractOperationSequences();
			Collection<OperationSequence> result = new ArrayList<OperationSequence>();

			Iterator<OperationSequence> iterator = c.iterator();
			Iterator<OperationSequence> iterator_2;

			OperationSequence os;
			OperationSequence os_temp;

			while (iterator.hasNext()) {

				iterator_2 = c2.iterator();
				os = iterator.next();

				while (iterator_2.hasNext()) {

					os_temp = new OperationSequence(os);

					os_temp.addOperationSequence(iterator_2.next());

					result.add(os_temp);

				}

			}

			return recursiveAbstractionStrictToAbstractOperationSequences(
					(i + 1), result);

		}

		// It isn't, let's just add it to every OperationSequence

		else {

			Iterator<OperationSequence> iterator = c.iterator();

			while (iterator.hasNext())
				iterator.next().addOperation(o);

			return recursiveAbstractionStrictToAbstractOperationSequences(
					(i + 1), c);

		}

	}

	/**
	 * Generates OperationSequence from the current one but without any
	 * abstractionStrictOperation Will modify the given Tree !
	 * 
	 * @return Collection of OperationSequence
	 */
	public Collection<OperationSequence> abstractionStrictToAbstractOperationSequences() {

		Collection<OperationSequence> c = new ArrayList<OperationSequence>();
		c.add(new OperationSequence());

		return recursiveAbstractionStrictToAbstractOperationSequences(0, c);

	}

	/**
	 * Method used to generate OperationSequences that isn't Abstract. It
	 * unabstract the Operation i from the current OperationSequence and make
	 * the cartesian product with c
	 * 
	 * @param i
	 *            Number of Operation to unabstract
	 * @param c
	 *            Collection that contains OperationSequence unabstracted given
	 *            by the i-x operations (x > 0)
	 * @param t
	 *            Tree to use to unabstract
	 * @return Cartesian product
	 */
	private Collection<OperationSequence> recursiveAbstractionToOperationSequences(
			int i, Collection<OperationSequence> c, Tree t)
			throws badOperationException, badAddressException, noNodeException,
			headerNodeException, addNodeException, removeNodeException,
			abstractOperationException {

		// If there isn't any operation at i

		if (i >= this.operations.size())
			return c;

		// Else get the operation

		Operation o = this.getOperation(i);

		// If it is abstract, let's make the cartesian product

		if (o instanceof UpdateOperation) {

			t = new Tree(t);

			UpdateOperation o2 = (UpdateOperation) o;

			Collection<OperationSequence> c2 = o2
					.abstractionToOperationSequences(t);
			Collection<OperationSequence> result = new ArrayList<OperationSequence>();

			Iterator<OperationSequence> iterator = c.iterator();
			Iterator<OperationSequence> iterator_2;

			OperationSequence os, temp_os;

			while (iterator.hasNext()) {

				os = new OperationSequence(iterator.next());
				iterator_2 = c2.iterator();

				while (iterator_2.hasNext()) {

					temp_os = new OperationSequence(os);
					temp_os.addOperationSequence(iterator_2.next());

					result.add(temp_os);

				}

			}

			return recursiveAbstractionToOperationSequences((i + 1), result, t);

		}

		// It's not abstract, we just have to add the operation to every
		// OperationSequence

		else {

			Iterator<OperationSequence> iterator = c.iterator();

			while (iterator.hasNext())
				iterator.next().addOperation(o);

			t = new Tree(t);
			o.make(t);

			return recursiveAbstractionToOperationSequences((i + 1), c, t);

		}

	}

	/**
	 * Generates OperationSequence from the current one but without any
	 * abstraction. Will modify the given Tree !
	 * 
	 * @param t
	 *            Tree to use to unabstract
	 * @return Collection of OperationSequence
	 */
	public Collection<OperationSequence> abstractionToOperationSequences(Tree t)
			throws badOperationException, badAddressException, noNodeException,
			headerNodeException, addNodeException, removeNodeException,
			abstractOperationException {

		// Creates the result

		Collection<OperationSequence> c = new ArrayList<OperationSequence>();

		// If the OS is abstract strict... Unabstract

		if (this.isAbstractStrict()) {

			Iterator<OperationSequence> i = this
					.abstractionStrictToAbstractOperationSequences().iterator();
			OperationSequence os;

			while (i.hasNext()) {

				os = i.next();
				c.addAll(os.abstractionToOperationSequences(t));

			}

		}

		// if it's just abstract

		else if (this.isAbstract()) {

			c.add(new OperationSequence());

			return recursiveAbstractionToOperationSequences(0, c, t);

		}

		// Else: nothing to do :)

		else {

			c.add(this);

		}

		// result

		return c;

	}

	/**
	 * Returns the OperationSequence formated to fit Xml as JDOM. Won't modifiy
	 * the given Tree
	 * 
	 * @param t
	 *            Tree to use for the operations
	 * @return An JDOM.Element
	 * @throws addNodeException
	 * @throws noNodeException
	 * @throws removeNodeException
	 * @throws badAddressException
	 * @throws headerNodeException
	 * @throws abstractOperationException
	 */
	public Element toXmlElement(Tree t, int id) throws addNodeException,
			noNodeException, removeNodeException, badAddressException,
			headerNodeException, abstractOperationException {

		t = new Tree(t);
		make(t);

		Element result = new Element("OperationSequence");

		Element tree = new Element("Tree");
		tree.addContent(t.toFlatString().replace("<", "[").replace(">", "]"));
		// tree.addContent(t.toString().replace("<", "[").replace(">", "]"));
		tree.setAttribute("ID", String.valueOf(id));

		result.addContent(tree);

		Iterator<Operation> i = this.operations.iterator();
		while (i.hasNext())
			result.addContent(i.next().toXmlElement());

		return result;

	}

	/**
	 * Returns the OperationSequence formated to fit Xml as JDOM. Won't modifiy
	 * the given Tree
	 * 
	 * @param t
	 *            Tree to use for the operations
	 * 
	 * @param vt
	 *            Vocabulary translator establishing the correspondence between
	 *            XML elements and characters
	 * 
	 * @return An JDOM.Element
	 * @throws addNodeException
	 * @throws noNodeException
	 * @throws removeNodeException
	 * @throws badAddressException
	 * @throws headerNodeException
	 * @throws abstractOperationException
	 */
	public Element toXmlElement(Tree t, VocabTranslator vt, int id)
			throws addNodeException, noNodeException, removeNodeException,
			badAddressException, headerNodeException,
			abstractOperationException {

		Element result = new Element("OperationSequence");

		Element tree = new Element("Tree");
		// tree.addContent(t.toString(vt).replace("<", "[").replace(">", "]"));
		tree.addContent(t.toFlatString(vt).replace("<", "[").replace(">", "]"));
		tree.setAttribute("ID", String.valueOf(id));

		result.addContent(tree);

		Iterator<Operation> i = this.operations.iterator();
		while (i.hasNext())
			result.addContent(i.next().toXmlElement(vt));

		return result;

	}

	/**
	 * Returns the OperationSequence formated to fit Xml as JDOM
	 * 
	 * @return JDOM Element
	 */
	public Element toXmlElement() {

		Element result = new Element("OperationSequence");

		Iterator<Operation> i = this.operations.iterator();
		while (i.hasNext())
			result.addContent(i.next().toXmlElement());

		return result;

	}

	/**
	 * Returns the OperationSequence formated to fit Xml as JDOM
	 * 
	 * @param vt
	 *            Vocabulary translator establishing the correspondence between
	 *            XML elements and characters
	 * 
	 * @return JDOM Element
	 */
	public Element toXmlElement(VocabTranslator vt) {

		Element result = new Element("OperationSequence");

		Iterator<Operation> i = this.operations.iterator();
		while (i.hasNext())
			result.addContent(i.next().toXmlElement(vt));

		return result;

	}

	/**
	 * Returns an iterator to access to every Operation of the OperationSequence
	 * 
	 * @return iterator
	 */
	public Iterator<Operation> iterator() {

		return operations.iterator();

	}

}
