/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.operations.abstractupdate;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import org.jdom.Element;

import fr.ufrt.xmlcorrector.exceptions.abstractOperationException;
import fr.ufrt.xmlcorrector.exceptions.addNodeException;
import fr.ufrt.xmlcorrector.exceptions.badAddressException;
import fr.ufrt.xmlcorrector.exceptions.badOperationException;
import fr.ufrt.xmlcorrector.exceptions.headerNodeException;
import fr.ufrt.xmlcorrector.exceptions.noMinimalTreeException;
import fr.ufrt.xmlcorrector.exceptions.noNodeException;
import fr.ufrt.xmlcorrector.exceptions.removeNodeException;
import fr.ufrt.xmlcorrector.operations.Operation;
import fr.ufrt.xmlcorrector.operations.OperationSequence;
import fr.ufrt.xmlcorrector.operations.update.AddTreeOperation;
import fr.ufrt.xmlcorrector.schema.TreeSchema;
import fr.ufrt.xmlcorrector.tree.Address;
import fr.ufrt.xmlcorrector.tree.Tree;
import fr.ufrt.xmlcorrector.vocabtranslator.VocabTranslator;

/**
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class AddMinimalTreeOperation extends AbstractUpdateOperation {

	Character header;
	TreeSchema ts;

	/**
	 * Constructor of the operations ReplaceTreeOperation
	 * 
	 * @param address
	 *            Address of the Node that should be replaced
	 * @param header
	 *            Character of the root of the Tree to create
	 * @param ts
	 *            TreeSchema to use to create the Minimal Tree
	 * @throws noMinimalTreeException
	 *             Throws this exception if there isn't any minimal tree for
	 *             this header
	 */
	public AddMinimalTreeOperation(Address address, Character header,
			TreeSchema ts) throws noMinimalTreeException {

		// Let's check if the TreeSchema is okay

		if (ts.getMinimalTrees(header) == null)
			throw new noMinimalTreeException();

		// It's okay

		this.address = address;
		this.header = header;
		this.ts = ts;
		this.operation = OperationType.addMinimalTree;

	}

	public Collection<OperationSequence> abstractionToOperationSequences(Tree t)
			throws badOperationException, badAddressException, noNodeException,
			headerNodeException, addNodeException, removeNodeException,
			abstractOperationException {

		// Creates the result

		Collection<OperationSequence> c = new ArrayList<OperationSequence>();

		// Creates the Collection that will contain AbstractOperationSequences
		// (but no strict)

		Collection<OperationSequence> abstract_c = abstractionStrictToAbstractOperationSequences();

		// For every abstract operation sequence

		Iterator<OperationSequence> i = abstract_c.iterator();

		while (i.hasNext()) {

			// Un abstract it and add it to the result

			c.addAll(i.next().abstractionToOperationSequences(t));

		}

		return c;

	}

	public Collection<OperationSequence> abstractionStrictToAbstractOperationSequences() {

		// Creates the result

		ArrayList<OperationSequence> sequences = new ArrayList<OperationSequence>();

		// Gets the Minimal Tree and for each...

		Iterator<Tree> iterator = ts.getMinimalTrees(this.header).iterator();
		OperationSequence os;

		while (iterator.hasNext()) {

			// Create an operation Sequence

			os = new OperationSequence();

			// Add an AddTreeOperation

			os.addOperation(new AddTreeOperation(this.address, iterator.next()));

			// Add the Sequence

			sequences.add(os);

		}

		return sequences;

	}

	public void make(Tree tree) throws abstractOperationException {

		throw new abstractOperationException();

	}

	public int getCost(Tree tree) {

		return ts.getMinimalTreeCost(this.header);

	}

	public int getCost() {

		return ts.getMinimalTreeCost(this.header);

	}

	public String toString() {

		return "(" + this.operation + ", \"" + this.address + "\", "
				+ this.header + ")";

	}

	public Operation clone() {

		try {
			return new AddMinimalTreeOperation(new Address(this.address),
					this.header, this.ts);
		} catch (noMinimalTreeException e) {
			return null;
		}

	}

	public boolean equals(Object o) {

		if (o instanceof AddMinimalTreeOperation)
			return o.hashCode() == this.hashCode();
		else
			return false;

	}

	public int hashCode() {

		return (this.operation + "o" + this.getAddress().getAddress() + "o" + this.header)
				.hashCode();

	}

	public Element toXmlElement() {

		Element result = new Element(this.operation.toString());
		result.setAttribute("Address", this.address.toString());
		result.setAttribute("Node", this.header.toString());

		return result;

	}

	/*
	 * @param vt Vocabulary translator establishing the correspondence between
	 * XML elements and characters
	 */
	public Element toXmlElement(VocabTranslator vt) {

		Element result = new Element(this.operation.toString());
		result.setAttribute("Address", this.address.toString());
		result.setAttribute("Node", vt.getElem(this.header));

		return result;

	}

}
