/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.operations.abstractupdate;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import org.jdom.Element;

import fr.ufrt.xmlcorrector.exceptions.abstractOperationException;
import fr.ufrt.xmlcorrector.exceptions.addNodeException;
import fr.ufrt.xmlcorrector.exceptions.badAddressException;
import fr.ufrt.xmlcorrector.exceptions.badOperationException;
import fr.ufrt.xmlcorrector.exceptions.headerNodeException;
import fr.ufrt.xmlcorrector.exceptions.noMinimalTreeException;
import fr.ufrt.xmlcorrector.exceptions.noNodeException;
import fr.ufrt.xmlcorrector.exceptions.removeNodeException;
import fr.ufrt.xmlcorrector.matrix.Result;
import fr.ufrt.xmlcorrector.operations.Operation;
import fr.ufrt.xmlcorrector.operations.OperationSequence;
import fr.ufrt.xmlcorrector.tree.Address;
import fr.ufrt.xmlcorrector.tree.Tree;
import fr.ufrt.xmlcorrector.vocabtranslator.VocabTranslator;

/**
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */

public class CorrectTreeOperation extends AbstractUpdateOperation {

	Character header;
	Result result;
	boolean minimal;

	/**
	 * Constructor of the operations CorrectTreeOperation when corrections are
	 * minimal
	 * 
	 * @param address
	 *            Address of the Node that should be replaced
	 * @param header
	 *            Character of the root of the Tree to create
	 * @param result
	 *            Result to use
	 * @throws noMinimalTreeException
	 *             Throws this exception if there isn't any minimal tree for
	 *             this header
	 */
	public CorrectTreeOperation(Address address, Character header, Result result)
			throws noMinimalTreeException {

		// It's okay

		this.address = address;
		this.header = header;
		this.result = result;
		this.operation = OperationType.CorrectTree;
		this.minimal = true;

	}

	/**
	 * Constructor of the operations CorrectTreeOperation when correction aren't
	 * minimal
	 * 
	 * @param address
	 *            Address of the Node that should be replaced
	 * @param header
	 *            Character of the root of the Tree to create
	 * @param result
	 *            Result to use
	 * @throws noMinimalTreeException
	 *             Throws this exception if there isn't any minimal tree for
	 *             this header
	 */
	public CorrectTreeOperation(Address address, Character header,
			Result result, boolean minimal) throws noMinimalTreeException {

		// It's okay

		this.address = address;
		this.header = header;
		this.result = result;
		this.operation = OperationType.CorrectTree;
		this.minimal = minimal;

	}

	public Collection<OperationSequence> abstractionStrictToAbstractOperationSequences() {

		// Create the result

		Collection<OperationSequence> result = new ArrayList<OperationSequence>();

		// For every Sequence of the Result

		Iterator<OperationSequence> i = this.result.getSequences().iterator();

		while (i.hasNext()) {

			// Get the sequence

			OperationSequence os = i.next();

			// Add an addressPrefix

			os.addAddressPrefix(this.address);

			// Add every unabstraction

			result.addAll(os.abstractionStrictToAbstractOperationSequences());

		}

		return result;

	}

	public Collection<OperationSequence> abstractionToOperationSequences(Tree t)
			throws badOperationException, badAddressException, noNodeException,
			headerNodeException, addNodeException, removeNodeException,
			abstractOperationException {

		// Create the result

		Collection<OperationSequence> c = new ArrayList<OperationSequence>();

		// Get the abstract sequences

		Collection<OperationSequence> abstract_c = abstractionStrictToAbstractOperationSequences();

		// For every sequence

		Iterator<OperationSequence> i = abstract_c.iterator();

		while (i.hasNext()) {

			// Add the un abstraction to the result

			c.addAll(i.next().abstractionToOperationSequences(t));

		}

		return c;

	}

	public void make(Tree tree) throws abstractOperationException {

		throw new abstractOperationException();

	}

	public int getCost(Tree tree) {

		return this.result.getCost();

	}

	public Operation clone() {

		try {

			return new CorrectTreeOperation(new Address(this.address),
					this.header, this.result, this.minimal);

		} catch (noMinimalTreeException e) {

			return null;

		}

	}

	public boolean equals(Object o) {

		if (o instanceof CorrectTreeOperation)
			return o.hashCode() == this.hashCode();
		else
			return false;

	}

	public int hashCode() {

		return (this.operation + "o" + this.getAddress().getAddress() + "o" + this.header)
				.hashCode();

	}

	public String toCompleteString() {

		String result = "(" + this.operation + ", \"" + this.address + "\", "
				+ this.header + "):" + this.result.getCost() + "("
				+ this.minimal + ")\n";
		result = result + "--------------\n";

		Iterator<OperationSequence> i = this.result.getSequences().iterator();
		while (i.hasNext()) {

			result = result + i.next() + "\n";

		}

		result = result + "--------------\n";

		return result;

	}

	public String toString() {

		return (this.minimal || this.result == null) ? "(" + this.operation
				+ ", \"" + this.address + "\", " + this.header + ")" : "("
				+ this.operation + ", \"" + this.address + "\", " + this.header
				+ ", " + this.result.getCost() + ")";

	}

	public Element toXmlElement() {

		Element result = new Element(this.operation.toString());
		result.setAttribute("Address", this.address.toString());
		result.setAttribute("Node", this.header.toString());

		return result;

	}

	/*
	 * @param vt Vocabulary translator establishing the correspondence between
	 * XML elements and characters
	 */

	public Element toXmlElement(VocabTranslator vt) {

		Element result = new Element(this.operation.toString());
		result.setAttribute("Address", this.address.toString());
		result.setAttribute("Node", vt.getElem(this.header));

		return result;

	}

}
