/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.operations.update;

import java.util.ArrayList;
import java.util.Collection;

import org.jdom.Element;

import fr.ufrt.xmlcorrector.exceptions.abstractOperationException;
import fr.ufrt.xmlcorrector.exceptions.addNodeException;
import fr.ufrt.xmlcorrector.exceptions.badAddressException;
import fr.ufrt.xmlcorrector.exceptions.badOperationException;
import fr.ufrt.xmlcorrector.exceptions.headerNodeException;
import fr.ufrt.xmlcorrector.exceptions.noNodeException;
import fr.ufrt.xmlcorrector.exceptions.removeNodeException;
import fr.ufrt.xmlcorrector.operations.Operation;
import fr.ufrt.xmlcorrector.operations.OperationSequence;
import fr.ufrt.xmlcorrector.tree.Address;
import fr.ufrt.xmlcorrector.tree.Tree;
import fr.ufrt.xmlcorrector.vocabtranslator.VocabTranslator;

/**
 * Add Tree Operation.
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class AddTreeOperation extends UpdateOperation {

	Tree tree;

	/**
	 * Constructor of the operations ReplaceTreeOperation
	 * 
	 * @param address
	 *            Address of the Node that should be replaced
	 * @param tree
	 *            Tree that should be used to replace
	 */
	public AddTreeOperation(Address address, Tree tree) {

		// It's okay

		this.address = address;
		this.tree = tree;
		this.operation = OperationType.addTree;

	}

	public Collection<OperationSequence> abstractionToOperationSequences(Tree t)
			throws badOperationException, badAddressException, noNodeException,
			headerNodeException, addNodeException, removeNodeException,
			abstractOperationException {

		// Gets the OperationSequence to create the tree

		OperationSequence os = this.tree.getCreationOperationsSequence();

		// Adds the address as prefix to every operation

		os.addAddressPrefix(this.address);

		// Makes the operation

		os.make(t);

		// Creates the result

		Collection<OperationSequence> c = new ArrayList<OperationSequence>();
		c.add(os);

		return c;

	}

	public void make(Tree tree) throws addNodeException, badAddressException {

		tree.addTree(this.address, this.tree);

	}

	public int getCost(Tree tree) {

		return this.tree.getCreationCost();

	}

	public String toString() {

		return "(" + this.operation + ", \"" + this.address + "\", "
				+ this.tree.toFlatString() + ")";

	}

	public Operation clone() {

		return new AddTreeOperation(new Address(this.address), new Tree(
				this.tree));

	}

	public boolean isAbstractStrict() {

		return false;

	}

	public boolean equals(Object o) {

		if (o instanceof AddTreeOperation)
			return o.hashCode() == this.hashCode();
		else
			return false;

	}

	public int hashCode() {

		return (this.operation + "o" + this.getAddress().getAddress() + "o" + this.tree
				.toFlatString()).hashCode();

	}

	public Element toXmlElement() {

		Element result = new Element(this.operation.toString());
		result.setAttribute("Address", this.address.toString());
		result.setAttribute("Tree", this.tree.toFlatString().replace("<", "[")
				.replace(">", "]"));

		return result;

	}

	public Element toXmlElement(VocabTranslator vt) {

		Element result = new Element(this.operation.toString());
		result.setAttribute("Address", this.address.toString());
		result.setAttribute("Tree", this.tree.toFlatString(vt)
				.replace("<", "[").replace(">", "]"));

		return result;

	}

}
