/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.operations.update;

import java.util.ArrayList;
import java.util.Collection;

import org.jdom.Element;

import fr.ufrt.xmlcorrector.exceptions.abstractOperationException;
import fr.ufrt.xmlcorrector.exceptions.addNodeException;
import fr.ufrt.xmlcorrector.exceptions.badAddressException;
import fr.ufrt.xmlcorrector.exceptions.badOperationException;
import fr.ufrt.xmlcorrector.exceptions.headerNodeException;
import fr.ufrt.xmlcorrector.exceptions.noNodeException;
import fr.ufrt.xmlcorrector.exceptions.removeNodeException;
import fr.ufrt.xmlcorrector.operations.Operation;
import fr.ufrt.xmlcorrector.operations.OperationSequence;
import fr.ufrt.xmlcorrector.tree.Address;
import fr.ufrt.xmlcorrector.tree.Tree;
import fr.ufrt.xmlcorrector.vocabtranslator.VocabTranslator;

/**
 * Remove tree Operation.
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class RemoveTreeOperation extends UpdateOperation {

	/**
	 * Constructor of the operation RemoveTreeOperation
	 * 
	 * @param address
	 *            Address of the sub Tree that has to be removed
	 */
	public RemoveTreeOperation(Address address) {

		// It's okay

		this.address = address;
		this.operation = OperationType.removeTree;

	}

	public Collection<OperationSequence> abstractionToOperationSequences(Tree t)
			throws badOperationException, badAddressException, noNodeException,
			addNodeException, removeNodeException, headerNodeException,
			abstractOperationException {

		// Gets the OperationSequence to remove the sub tree

		OperationSequence os = t.getNodeAsTree(this.address)
				.getRemoveOperationsSequence();

		// Adds the Address as prefix of every Operation

		os.addAddressPrefix(this.address);

		// Makes the operation

		os.make(t);

		// Adds to the result

		Collection<OperationSequence> c = new ArrayList<OperationSequence>();
		c.add(os);

		return c;

	}

	public int getCost(Tree tree) throws noNodeException {

		Tree t = tree.getNodeAsTree(this.address);

		return t.getCreationCost();

	}

	public void make(Tree tree) throws noNodeException, badAddressException {

		tree.removeTree(this.address);

	}

	public String toString() {

		return "(" + this.operation + ", \"" + this.address + "\")";

	}

	public Operation clone() {

		return new RemoveTreeOperation(new Address(this.address));

	}

	public boolean isAbstractStrict() {

		return false;

	}

	public boolean equals(Object o) {

		if (o instanceof RemoveTreeOperation)
			return o.hashCode() == this.hashCode();
		else
			return false;

	}

	public int hashCode() {

		return (this.operation + "o" + this.getAddress().getAddress())
				.hashCode();

	}

	public Element toXmlElement() {

		Element result = new Element(this.operation.toString());
		result.setAttribute("Address", this.address.toString());

		return result;

	}

	public Element toXmlElement(VocabTranslator vt) {

		Element result = new Element(this.operation.toString());
		result.setAttribute("Address", this.address.toString());

		return result;

	}

}
