/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.operations.update;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import org.jdom.Element;

import fr.ufrt.xmlcorrector.exceptions.abstractOperationException;
import fr.ufrt.xmlcorrector.exceptions.addNodeException;
import fr.ufrt.xmlcorrector.exceptions.badAddressException;
import fr.ufrt.xmlcorrector.exceptions.badOperationException;
import fr.ufrt.xmlcorrector.exceptions.headerNodeException;
import fr.ufrt.xmlcorrector.exceptions.noNodeException;
import fr.ufrt.xmlcorrector.exceptions.removeNodeException;
import fr.ufrt.xmlcorrector.matrix.Result;
import fr.ufrt.xmlcorrector.operations.Operation;
import fr.ufrt.xmlcorrector.operations.OperationSequence;
import fr.ufrt.xmlcorrector.tree.Address;
import fr.ufrt.xmlcorrector.tree.Tree;
import fr.ufrt.xmlcorrector.vocabtranslator.VocabTranslator;

/**
 * Replace Tree Operation.
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class ReplaceTreeOperation extends UpdateOperation {

	Tree tree;
	Result result;

	/**
	 * Constructor of the operations ReplaceTreeOperation
	 * 
	 * @param address
	 *            Address of the Node that should be replaced
	 * @param tree
	 *            Tree that should be used to replace
	 */
	public ReplaceTreeOperation(Address address, Tree tree, Result result) {

		// It's okay

		this.address = address;
		this.tree = tree;
		this.operation = OperationType.replaceTree;
		this.result = result;

	}

	public Collection<OperationSequence> abstractionToOperationSequences(Tree t)
			throws badOperationException, badAddressException, noNodeException,
			headerNodeException, addNodeException, removeNodeException,
			abstractOperationException {

		// Creates the result

		Collection<OperationSequence> c = new ArrayList<OperationSequence>();

		// For every OperationSequence of the Result

		Iterator<OperationSequence> i = this.result.getSequences().iterator();

		while (i.hasNext()) {

			// Get the OperationSequence

			OperationSequence os_temp = i.next();

			// Adds the address as prefix to every Operation

			os_temp.addAddressPrefix(this.address);

			// For every non abstract OperationSequence generated by os_temp

			Iterator<OperationSequence> iterator = os_temp
					.abstractionToOperationSequences(t).iterator();

			while (iterator.hasNext()) {

				// Get it

				OperationSequence os = iterator.next();

				// Add it

				c.addAll(os.abstractionToOperationSequences(t));

			}

		}

		// Make it to the tree

		this.make(t);

		return c;

	}

	public void make(Tree tree) throws noNodeException, addNodeException,
			removeNodeException, badAddressException {

		tree.replaceTree(this.address, this.tree);

	}

	public int getCost(Tree tree) {

		return this.result.getCost();

	}

	public String toString() {

		return "(" + this.operation + ", \"" + this.address + "\", "
				+ this.tree.toFlatString() + ")";

	}

	public String toCompleteString() {

		String result = "(" + this.operation + ", \"" + this.address + "\", "
				+ this.tree.toFlatString() + ")\n";

		result = result + "--------------\n";

		Iterator<OperationSequence> i = this.result.getSequences().iterator();
		while (i.hasNext()) {

			result = result + i.next() + "\n";

		}

		result = result + "--------------\n";

		return result;

	}

	public Operation clone() {

		return new ReplaceTreeOperation(new Address(this.address), new Tree(
				this.tree), new Result(this.result));

	}

	public boolean isAbstractStrict() {

		return false;

	}

	public boolean equals(Object o) {

		if (o instanceof ReplaceTreeOperation)
			return o.hashCode() == this.hashCode();
		else
			return false;

	}

	public int hashCode() {

		return (this.operation + "o" + this.getAddress().getAddress() + "o" + this.tree
				.toFlatString()).hashCode();

	}

	public Element toXmlElement() {

		Element result = new Element(this.operation.toString());
		result.setAttribute("Address", this.address.toString());
		result.setAttribute("Node", this.tree.toFlatString().replace("<", "[")
				.replace(">", "]"));

		return result;

	}

	public Element toXmlElement(VocabTranslator vt) {

		Element result = new Element(this.operation.toString());
		result.setAttribute("Address", this.address.toString());
		result.setAttribute("Node", this.tree.toFlatString(vt)
				.replace("<", "[").replace(">", "]"));

		return result;

	}

}
