/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.schema;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;
import java.util.Map.Entry;

import dk.brics.automaton.Automaton;
import dk.brics.automaton.RegExp;
import fr.ufrt.xmlcorrector.exceptions.addNodeException;
import fr.ufrt.xmlcorrector.exceptions.addSchemaException;
import fr.ufrt.xmlcorrector.exceptions.badAddressException;
import fr.ufrt.xmlcorrector.exceptions.badSchemaException;
import fr.ufrt.xmlcorrector.exceptions.generatedSchemaException;
import fr.ufrt.xmlcorrector.exceptions.generationSchemaException;
import fr.ufrt.xmlcorrector.exceptions.headerNodeException;
import fr.ufrt.xmlcorrector.exceptions.onlyCharactersException;
import fr.ufrt.xmlcorrector.tree.Address;
import fr.ufrt.xmlcorrector.tree.Node;
import fr.ufrt.xmlcorrector.tree.Tree;

/**
 * Class TreeSchema.
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class TreeSchema {

	Character root;
	HashMap<Character, Automaton> automatas;
	HashMap<Character, String> regexps;
	HashMap<Character, ArrayList<Tree>> minimal_trees;
	HashMap<Character, Integer> minimal_costs;
	boolean minimal_generated = false;

	// HashMap<Character, ArrayList<Tree>> underthreshold_tree;
	// HashMap<Character, Integer> underthresholdminimal_costs;
	// boolean underthreshold_generated = false;

	/**
	 * Constructor that creates a new TreeSchema where the root node should be
	 * named like c
	 * 
	 * @param c
	 *            Header of the root
	 */
	public TreeSchema(Character c) {

		this.root = c;
		this.automatas = new HashMap<Character, Automaton>();
		this.regexps = new HashMap<Character, String>();
		this.minimal_trees = new HashMap<Character, ArrayList<Tree>>();
		this.minimal_costs = new HashMap<Character, Integer>();
		// this.underthreshold_tree = new HashMap<Character, ArrayList<Tree>>();
		// this.underthresholdminimal_costs = new HashMap<Character, Integer>();

	}

	/**
	 * Adds an expreg for the character given. It is considered as a Schema and
	 * create an Automaton
	 * 
	 * @param element
	 *            Header of the elements that should satisfy the expreg
	 * @param expreg
	 *            ExpReg of every element
	 * @throws addSchemaException
	 *             Throws this Exception if the TreeSchema contains already a
	 *             Schema for this header
	 * @throws generatedSchemaException
	 *             Throws this Exception the minimal trees of the TreeSchema
	 *             have already been calculated
	 * @throws badSchemaException
	 *             Throws this Exception if the Automaton given by the ExpReg
	 *             cannot be satisfied
	 */
	public void addSchema(Character element, String expreg)
			throws addSchemaException, generatedSchemaException,
			badSchemaException {

		if (automatas.containsKey(element))
			throw new addSchemaException();
		else if (minimal_generated)
			throw new generatedSchemaException();

		if (expreg != "") {

			RegExp r = new RegExp(expreg);
			Automaton a = r.toAutomaton();

			if (a.isEmpty())
				throw new badSchemaException();

			automatas.put(element, a);
			this.regexps.put(element, expreg);

		} else {

			this.automatas.put(element, null);
			this.regexps.put(element, "empty");

		}

	}

	/**
	 * Returns true if the Node is locally valid: it doesn't check the validity
	 * of the children<br />
	 * If there isn't any expreg for the given Header, the node should be a leaf
	 * to be valid
	 * 
	 * @param node
	 *            Node that should be checked
	 * @return validity (true)
	 * @throws onlyCharactersException
	 *             Throws this Exception if the node hasn't a compatible Header
	 */
	public boolean isValid(Node node) throws onlyCharactersException {

		boolean result = false;

		// He does, let's take it

		Character c = node.getHeader();

		// Okay now let's look if we have a rule for him

		if (this.automatas.containsKey(c) && this.automatas.get(c) != null)
			result = this.automatas.get(c).run(node.getChildrenAsString());

		// No rule, if he's a leaf it's okay i think

		else
			result = node.isLeaf();

		return result;

	}

	/**
	 * Checks if the Node is valid and if its children are also<br />
	 * It checks the current node using isValid
	 * 
	 * @param node
	 *            Node to check
	 * @return True if the Node and its children are valid
	 * @throws onlyCharactersException
	 *             Throws this Exception if the Header of a node isn't
	 *             compatible
	 */
	public boolean isDeepValid(Node node) throws onlyCharactersException {

		if (!isValid(node))
			return false;
		else {

			for (int i = 0; i < node.getChildrenNumber(); i++) {

				if (!isDeepValid(node.getChild(i)))
					return false;

			}

			return true;

		}

	}

	/**
	 * Checks if a Tree is valid by using isDeepValid on the root
	 * 
	 * @param tree
	 *            Tree to check
	 * @return Returns true if the Tree is valid
	 * @throws onlyCharactersException
	 *             Throws this Exception if there is a Node with an uncompatible
	 *             Header
	 */
	public boolean isValid(Tree tree) throws onlyCharactersException {

		Node root = tree.getRoot();

		if (root == null)
			return false;
		else if (root.getHeader() == this.root && isDeepValid(root))
			return true;
		else
			return false;

	}

	/**
	 * Returns the cost of creation of the minimal tree for the given Header
	 * 
	 * @param c
	 *            Header of the minimal tree to calculate
	 * @return cost of creation
	 */
	public int getMinimalTreeCost(Character c) {

		if (!this.minimal_costs.containsKey(c))
			return -1;
		else
			return this.minimal_costs.get(c);

	}

	/**
	 * Returns an ArrayList containing every MinimalTree for the given Header
	 * 
	 * @param c
	 *            Header of the root of the minimal tree
	 * @return minimal trees
	 */
	public ArrayList<Tree> getMinimalTrees(Character c) {

		if (!this.minimal_trees.containsKey(c))
			return null;
		else
			return this.minimal_trees.get(c);

	}

	/**
	 * Generation of every combination possible
	 * 
	 * @param list
	 *            List of the Tree to use to generate the combinations
	 * @param root_tree
	 *            Tree to use to add a sub tree
	 * @return the combinations
	 * @throws addNodeException
	 *             Throws this Exception if there is an error while creating
	 *             node
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while create
	 *             Addresses
	 */
	private static ArrayList<Tree> generateEveryTree(
			ArrayList<ArrayList<Tree>> list, Tree root_tree, int max_cost)
			throws addNodeException, badAddressException {

		ArrayList<Tree> results = new ArrayList<Tree>();

		// the list isn't null
		if (list == null || list.size() <= 0)
			return results;

		// let's get the first ArrayList
		ArrayList<Tree> current = list.get(0);

		// is the list multiple ?
		if (list.size() > 1) {

			list.remove(0);
			Iterator<Tree> i = current.iterator();
			Iterator<Tree> j;
			ArrayList<Tree> recursive_results = generateEveryTree(list,
					root_tree, max_cost);
			Tree t, v;

			while (i.hasNext()) {

				t = i.next();
				j = recursive_results.iterator();

				while (j.hasNext()) {

					v = new Tree(j.next());
					if (v.getCreationCost() + t.getCreationCost() <= max_cost) {
						v.addTree(new Address("0"), t);
						results.add(v);
					}

				}

			}

		}

		// no, there is a single list
		else {

			Iterator<Tree> i = current.iterator();
			Tree t, v;

			while (i.hasNext()) {

				t = new Tree(root_tree);

				v = new Tree(i.next());
				if (v.getCreationCost() + t.getCreationCost() <= max_cost) {
					t.addTree(new Address("0"), v);
					results.add(t);
				}

			}

		}

		return results;

	}

	/**
	 * 
	 * @param c
	 *            Character for wich we want to generate the minimal tree
	 * @param a
	 *            Automaton of the Character
	 * @param max_cost
	 *            Maximum cost that could have the minimal tree
	 * @param current_cost
	 *            Current cost for the recursive version
	 * @return A vector containing the trees generated, the cost, and if the
	 *         current_cost has played
	 * @throws generationSchemaException
	 *             Throws this Exception if there is an error while generating a
	 *             Schema
	 * @throws addNodeException
	 *             Throws this Exception if there an error while creating node
	 * @throws badAddressException
	 *             Throws this Exception if there an error with Addresses
	 * @throws headerNodeException
	 *             Throws this exception is the header of the new node is ""
	 */
	@SuppressWarnings("unchecked")
	private Vector generateMinimalTree(Character c, Automaton a,
			Integer max_cost, Integer current_cost, boolean minimal)
			throws generationSchemaException, addNodeException,
			badAddressException, headerNodeException {

		ArrayList<Tree> trees = new ArrayList<Tree>();

		Vector resultat = new Vector();

		int cost = 1;
		boolean cost_limitation = false;

		if (a != null) {

			// only if the empty string doesn't works :x

			if ((minimal && !a.run("")) || !minimal) {

				if (!minimal && a.run(""))
					trees.add(new Tree(c));
				// Ok so we want to get the minimal Tree, but maybe one day we
				// will have to get every minimal tree..
				// so we use the method getStrings(length) on our automaton
				// this method gives us every String accepted that is no longer
				// than length
				// we start with i = 0 to i...

				ArrayList<String> tree_expressions = new ArrayList<String>();
				Set<String> results = null;
				int length = 1;

				while (length <= max_cost) {

					results = a.getStrings(length);

					if (results != null && results.size() > 0) {

						tree_expressions.addAll(results);

					}

					length++;

				}

				// We found something ?

				if (tree_expressions.size() > 0) {

					// Ok now we have the Set of the Strings
					// We have to build a Set of Costs and chose the expression
					// we want... damn complicated.

					Iterator<String> iterator = tree_expressions.iterator();

					ArrayList<ArrayList<Tree>> minimal_trees_foreach_node;
					Iterator<Tree> minimal_trees_iterator;
					ArrayList<Tree> minimal_trees_generated;

					Vector recursive_result;
					ArrayList<Tree> recursive_trees;

					ArrayList<Tree> possible_minimal_trees = new ArrayList<Tree>();
					ArrayList<Integer> possible_minimal_trees_cost = new ArrayList<Integer>();

					String tree_expression;
					Character header;
					int count, total_temp_cost, temp_cost;
					boolean expression_valid = true;

					// Ok so let's see EVERY string that is possibly a solution
					// <= cost
					// go

					while (iterator.hasNext()) {

						// Get the expression
						tree_expression = iterator.next();

						// If the expression contains the current header, it's
						// no use to go further
						if (tree_expression.indexOf(c.toString()) >= 0)
							continue;

						// reinitialise the cost
						total_temp_cost = cost;

						// Currently, the expression is valid
						expression_valid = true;

						// Intialize the lists of possibilities
						minimal_trees_foreach_node = new ArrayList<ArrayList<Tree>>();

						// Ok so let's add every Character ... and get their
						// minimal trees
						for (count = 0; count < tree_expression.length(); count++) {

							header = tree_expression.charAt(count);

							if (!this.minimal_trees.containsKey(header)
									&& (total_temp_cost + current_cost) <= max_cost) {

								if (this.automatas.containsKey(header)) {

									recursive_result = minimal ? generateMinimalTree(
											header, this.automatas.get(header),
											max_cost, total_temp_cost
													+ current_cost, minimal)
											: generateMinimalTree(header,
													this.automatas.get(header),
													max_cost, 0, minimal);

								} else {

									recursive_result = generateMinimalTree(
											header, null, max_cost, 0, minimal);

								}

								if (recursive_result == null) {

									expression_valid = false;
									break;

								} else {

									recursive_trees = (ArrayList<Tree>) recursive_result
											.get(0);
									temp_cost = (Integer) recursive_result
											.get(1);

									if (!cost_limitation && current_cost > 0)
										cost_limitation = (Boolean) recursive_result
												.get(2);

								}

							} else if ((total_temp_cost + current_cost) <= max_cost) {

								recursive_trees = this.minimal_trees
										.get(header);
								temp_cost = this.minimal_costs.get(header);

								if (recursive_trees == null) {

									expression_valid = false;
									break;

								}

							} else {

								cost_limitation = true;
								expression_valid = false;
								break;

							}

							if ((total_temp_cost + temp_cost + current_cost) > max_cost
									|| temp_cost < 0
									|| recursive_trees.size() == 0) {

								expression_valid = false;
								break;

							}

							total_temp_cost = total_temp_cost + temp_cost;
							minimal_trees_foreach_node.add(recursive_trees);

						}

						// this solution was too expensive ? try next
						if (!expression_valid)
							continue;

						// Okay so let's generate every possibility now

						minimal_trees_generated = generateEveryTree(
								minimal_trees_foreach_node, new Tree(c),
								max_cost);

						if (minimal_trees_generated != null) {

							minimal_trees_iterator = minimal_trees_generated
									.iterator();

							// we've got every possibilities, but they aren't
							// the minimal ones for sure...

							while (minimal_trees_iterator.hasNext()) {

								possible_minimal_trees
										.add(minimal_trees_iterator.next());
								possible_minimal_trees_cost
										.add(total_temp_cost);

							}

						}

					}

					// Okay we generated every possible minimal tree... But
					// that's not what we want.
					// We want THE minimal tree.

					// So let's go check the minimal cost

					int minimal_cost = -1;

					for (count = 0; count < possible_minimal_trees_cost.size(); count++) {

						if (possible_minimal_trees_cost.get(count) < minimal_cost
								|| minimal_cost == -1)
							minimal_cost = possible_minimal_trees_cost
									.get(count);

					}

					if (minimal) {
						// Ok let's get every tree with this cost :)

						for (count = 0; count < possible_minimal_trees_cost
								.size(); count++) {

							if (possible_minimal_trees_cost.get(count) == minimal_cost)
								trees.add(possible_minimal_trees.get(count));

						}
					} else
						trees.addAll(possible_minimal_trees);

					// ok let's define the cost

					cost = minimal_cost;

				}

				else if (current_cost > 0)
					return null;

			} else
				trees.add(new Tree(c));

		} else
			trees.add(new Tree(c));

		if (!cost_limitation && (trees.size() > 0 || current_cost == 0)) {

			if (trees.size() <= 0) {

				this.minimal_trees.put(c, null);
				this.minimal_costs.put(c, -1);

			}

			else {

				this.minimal_trees.put(c, trees);
				this.minimal_costs.put(c, cost);

			}

		}

		resultat.add(trees);
		resultat.add(cost);
		resultat.add(cost_limitation);

		return resultat;

	}

	/**
	 * Generates minimal trees
	 * 
	 * @param max_cost
	 *            Maximum cost of a minimal tree
	 * @throws generatedSchemaException
	 *             Throws this Exception if the minimal trees have already been
	 *             generateed
	 * @throws generationSchemaException
	 *             Throws this Exception if there is an error while generating
	 * @throws addNodeException
	 *             Throws this Exception if there is an error while creating
	 *             nodes
	 * @throws badAddressException
	 *             Throws this Exception if there is an error in the addresses
	 * @throws headerNodeException
	 *             Throws this exception is the header of the new node is ""
	 */
	public void generateMinimalTrees(int max_cost, boolean minimal)
			throws generatedSchemaException, generationSchemaException,
			addNodeException, badAddressException, headerNodeException {

		if (minimal_generated)
			throw new generatedSchemaException();

		Iterator<Character> i = automatas.keySet().iterator();
		Character c;

		while (i.hasNext()) {

			c = i.next();

			if (!minimal_trees.containsKey(c)) {

				generateMinimalTree(c, automatas.get(c), max_cost, 0, minimal);

			}

		}

		minimal_generated = true;

	}

	public String toString() {

		String resultat = "----------------------------------\n";
		resultat = resultat + "Schema:\n";

		resultat = resultat + "\t Root: " + this.root + "\n";

		Iterator<Entry<Character, String>> i = this.regexps.entrySet()
				.iterator();
		Entry<Character, String> e;

		while (i.hasNext()) {

			e = i.next();
			resultat = resultat + "\t " + e.getKey() + ": " + e.getValue()
					+ "\n";

		}

		if (minimal_generated) {

			resultat = resultat + "----------------------------------\n";
			resultat = resultat + "Minimal Trees:\n";

			Iterator<Entry<Character, ArrayList<Tree>>> j = this.minimal_trees
					.entrySet().iterator();
			Entry<Character, ArrayList<Tree>> t;
			int c = 0;

			while (j.hasNext()) {

				t = j.next();
				resultat = resultat + "----------------------------------\n";
				resultat = resultat + "\t Minimal trees for " + t.getKey()
						+ " " + this.getMinimalTreeCost(t.getKey()) + ": \n";
				resultat = resultat + "----------------------------------\n";

				if (t.getValue() != null) {

					for (c = 0; c < t.getValue().size(); c++) {

						resultat = resultat + t.getValue().get(c) + "\n";

					}

				} else {

					resultat = resultat + "no minimal tree\n";

				}
			}

		}

		return resultat;

	}

	/**
	 * 
	 * @return the root of the TreeSchema
	 */
	public char getRoot() {

		return this.root;

	}

	/**
	 * Return the automaton corresponding to the tag c
	 * @param c
	 * @return automaton
	 */
	public Automaton getAutomaton(char c) {

		return this.automatas.get(c);

	}

	/**
	 * Compute the cost of the minimal tree having the tag c as root.
	 * @param c
	 * @return minimal cost
	 */
	public int getPossibleCostMinimalTree(char c) {
		int cost = 1;
		Automaton a = automatas.get(c);

		if (a != null) {
			String s = a.getShortestExample(true);

			for (int i = 0; i < s.length(); i++) {
				cost = cost + getPossibleCostMinimalTree(s.charAt(0));
			}
		}

		return cost;
	}

}
