/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.tree;

import fr.ufrt.xmlcorrector.exceptions.badAddressException;

/**
 * Class Node Address.
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class Address {

	String address;

	/**
	 * Creates an instance of Address using the String given<br />
	 * WARNING: if the adress isn't well formed, it will take the "0" address.<br />
	 * An address should be like "0.0", "0.1", "0", "1.1", but no "01.1", etc.<br />
	 * An address has to respect this expreg:
	 * ([0-9]|[1-9][0-9]*)(\.([0-9]|[1-9][0-9]*))*
	 * 
	 * @param address
	 *            XML address or position of a node
	 * @throws badAddressException
	 *             If the String of the address isn't valid
	 */
	public Address(String address) throws badAddressException {

		if (address.matches("([0-9]|[1-9][0-9]*)(\\.([0-9]|[1-9][0-9]*))*"))
			this.address = address;
		else if (address == "")
			this.address = address;
		else
			throw new badAddressException();

	}

	/**
	 * Creates an instance of Address by copying the given address
	 * 
	 * @param address
	 *            Given Address to copy
	 */
	public Address(Address address) {

		String address_string = address.getAddress();
		this.address = address_string;

	}

	/**
	 * Changes the start of the address by using the given Address and adding it
	 * to the current one.
	 * 
	 * @param prefix
	 *            Address that has to be used as prefix
	 */
	public void addPrefix(Address prefix) {

		this.address = (this.isRoot()) ? prefix.getAddress() : prefix
				.getAddress() + "." + this.address;

	}

	/**
	 * Returns the address as a string
	 * 
	 * @return address
	 */
	public String getAddress() {

		return this.address;

	}

	/**
	 * Returns a table of String<br />
	 * for each one, it is the local position <br />
	 * example: 0.0.1 gives "0", "0", "1"
	 * 
	 * @return Table of positions
	 */
	public String[] getPositions() {

		String[] result;

		// If we are at the root

		if (this.address.indexOf('.') == -1) {

			result = new String[1];
			result[0] = this.address;

		}

		// Else we split

		else {

			result = this.address.split("\\.");
			;

		}

		return result;

	}

	/**
	 * Returns the address of the father if it exists
	 * 
	 * @return address of the father or null
	 * @throws badAddressException
	 */
	public Address getFatherAddress() throws badAddressException {

		if (this.address.indexOf('.') == -1)
			return new Address("");
		else {

			String[] addresses = this.address.split("\\.");
			String father_address = "";

			for (int i = 0; i < addresses.length - 1; i++) {

				if (father_address != "")
					father_address = father_address + ".";
				father_address = father_address + addresses[i];

			}

			return new Address(father_address);

		}
		/**
		 * Correction int i = this.address.lastIndexOf("."); int j = (i == -1) ?
		 * 0 : i ; return new Address(this.address.substring(0, j)) ;
		 */

	}

	/**
	 * Returns the Address of the Child at the given position
	 * 
	 * @param i
	 *            position of the Child
	 * @return Address of the Child
	 * @throws badAddressException
	 */
	public Address getChildAddress(int i) throws badAddressException {

		return (!this.isRoot()) ? new Address(this.address + "." + i)
				: new Address(String.valueOf(i));

	}

	/**
	 * Returns the local position of the Node at the Address <br />
	 * Ex: 0.5.4 gives 4
	 * 
	 * @return local position
	 */
	public Integer getPosition() {

		if (this.address.indexOf('.') == -1)
			return Integer.valueOf(this.address);
		else {

			String[] addresses = this.address.split("\\.");
			return Integer.valueOf(addresses[(addresses.length - 1)]);

		}

		/**
		 * Correction int i = this.address.lastIndexOf("."); int j = (i == -1) ?
		 * 0 : i ; return Integer.valueOf(this.address.substring(j,
		 * this.address.length)) ;
		 */

	}

	/**
	 * Returns true if the Address is the Address of the Root
	 * 
	 * @return True if it is a Root Address
	 */
	public boolean isRoot() {

		return this.address == "";

	}

	/**
	 * Looks if the two address are same level (exemple: 0.0.1 and 0.0.2 but not
	 * 0.1.1 and 0.0.1 or 0.1.1.0 and 0.1.1)
	 * 
	 * @param a
	 *            Address to compare
	 * @return boolean
	 */
	public boolean isSameLevel(Address a) {

		// There is a root ?
		if (this.isRoot() || a.isRoot()) {

			return (this.isRoot() && a.isRoot()) ? true : false;

		} else {

			String[] addresses_this = this.address.split("\\.");
			String[] addresses_other = a.address.split("\\.");

			// if there isn't at a similar level, no need to verify
			if (addresses_this.length != addresses_other.length)
				return false;

			// Let's go down

			int i = 0;

			while (i < (addresses_this.length - 1)) {

				if (!addresses_this[i].equals(addresses_other[i]))
					return false;
				i++;

			}

			// Okay !
			return true;
			/*
			 * Correction if(this.isRoot() && a.isRoot()) return true ; else
			 * return this.getFatherAddress().equals(a.getFatherAddress()));
			 */

		}

	}

	public boolean equals(Object address) {

		if (address instanceof Address) {

			Address a = (Address) address;

			return (this.address.equals(a.getAddress()));

		}

		else
			return false;

	}

	public int hashCode() {

		return address.hashCode();

	}

	public String toString() {

		return address;

	}

}
