/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.tree;

import java.util.ArrayList;
import fr.ufrt.xmlcorrector.exceptions.noNodeException;

/**
 * Class Node.
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class Node {

	Character header;
	ArrayList<Node> children;

	/**
	 * Constructor using a Node using a Character converted as a String as
	 * Header
	 * 
	 * @param header
	 *            Character to convert and use as Header
	 */
	public Node(Character header) {

		this.header = header;
		this.children = new ArrayList<Node>();

	}

	/**
	 * Constructor using a Node to create a new One
	 * 
	 * @param source
	 *            Node to copy
	 */
	public Node(Node source) {

		this.header = source.getHeader();
		this.children = new ArrayList<Node>();

		for (int i = 0; i < source.getChildrenNumber(); i++) {

			addChild(i, new Node(source.getChild(i)));

		}

	}

	/**
	 * Deletes the Node at position i
	 * 
	 * @param i
	 *            Position of the Node to delete
	 * @throws noNodeException
	 *             Throws this Exception if the Node to delete doesn't exist
	 */
	public void removeChild(int i) throws noNodeException {

		if (i >= 0 && i < getChildrenNumber()) {

			this.children.remove(i);

		} else
			throw new noNodeException();

	}

	/**
	 * Add a child to the Node at the position i
	 * 
	 * @param i
	 *            Position where the Node should be added
	 * @param n
	 *            Node to add
	 */
	public void addChild(int i, Node n) {

		this.children.add(i, new Node(n));

	}

	/**
	 * Checks if the Node has children
	 * 
	 * @return true if the Node has children, false if not
	 */
	public boolean isLeaf() {

		return (getChildrenNumber() == 0);

	}

	/**
	 * Returns direct access to the Child at position i
	 * 
	 * @param i
	 *            Position of the Node to return
	 * @return Node The child if it exists, null in other case
	 */
	public Node getChild(int i) {

		return (i >= 0 && i < getChildrenNumber()) ? this.children.get(i)
				: null;

	}

	/**
	 * Returns the String that caracterize the direct children <br />
	 * For exemple, <a><b></b><c></c></a> gives "bc"
	 * 
	 * @return String
	 */
	public String getChildrenAsString() {

		String result = "";

		if (!isLeaf()) {

			for (int i = 0; i < getChildrenNumber(); i++) {

				result = result + getChild(i).getHeader();

			}

		}

		return result;

	}

	/**
	 * Changes the Header of the Node with the given Character that will be
	 * converted to a String
	 * 
	 * @param header
	 *            Character to use to change the Header
	 */
	public void setHeader(Character header) {

		this.header = header;

	}

	/**
	 * Returns the Header of the Node
	 * 
	 * @return Header of the Node
	 */
	public Character getHeader() {

		return this.header;

	}

	/**
	 * Gives the number of children that possess the Node
	 * 
	 * @return number of children
	 */
	public int getChildrenNumber() {

		return this.children.size();

	}

	/**
	 * Returns the Node's Header
	 */
	public String toString() {

		return String.valueOf(this.header);

	}

}
