/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.tree;

import java.io.File;
import java.util.Iterator;
import org.jdom.*;
import org.jdom.input.*;

import fr.ufrt.xmlcorrector.exceptions.VocabTranslatorException;
import fr.ufrt.xmlcorrector.exceptions.addNodeException;
import fr.ufrt.xmlcorrector.exceptions.badAddressException;
import fr.ufrt.xmlcorrector.exceptions.badOperationException;
import fr.ufrt.xmlcorrector.exceptions.headerNodeException;
import fr.ufrt.xmlcorrector.exceptions.noNodeException;
import fr.ufrt.xmlcorrector.exceptions.removeNodeException;
import fr.ufrt.xmlcorrector.exceptions.xmlVocabularySizeTooBig;
import fr.ufrt.xmlcorrector.operations.OperationSequence;
import fr.ufrt.xmlcorrector.operations.edit.AddNodeOperation;
import fr.ufrt.xmlcorrector.operations.edit.RemoveNodeOperation;
import fr.ufrt.xmlcorrector.vocabtranslator.VocabTranslator;

/**
 * Property editor for locales.
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class Tree {

	Node root;

	/**
	 * Constructor that creates an empty Tree
	 */
	public Tree() {

		this.root = null;

	}

	/**
	 * Constructor using the parameter to create a Node that will be the root
	 * 
	 * @param root
	 *            String that will be the Label of the root Node
	 * @throws headerNodeException
	 *             Throws this exception is the header of the new node is ""
	 */
	public Tree(Character root) throws headerNodeException {

		this.root = new Node(root);

	}

	/**
	 * Constructor using the parameter as the root
	 * 
	 * @param root
	 *            Node that will be the root of the new tree
	 */
	private Tree(Node root) {

		this.root = new Node(root);

	}

	/**
	 * Constructor that copies the given Tree
	 * 
	 * @param tree
	 *            Tree that should be copied
	 */
	public Tree(Tree tree) {

		if (!tree.isEmpty())
			root = new Node(tree.getRoot());
		else
			root = null;

	}

	/**
	 * Constructor building a Tree with a XML document (File) <br />
	 * uses JDOM to parse and build the tree
	 * 
	 * @param file
	 *            File to parse
	 * @throws Exception
	 *             Throws exception if the document isn't valid or if the SAX
	 *             Parser can't parse it
	 */
	@SuppressWarnings("unchecked")
	public Tree(File file) throws Exception {

		SAXBuilder sxb = new SAXBuilder();

		org.jdom.Document document;
		document = sxb.build(file);

		Element root = document.getRootElement();

		// Si il y a bien une racine
		if (root != null) {

			// Racine qui d�nome l'arbre vide ?

			if (root.getName().equals("EmptyTree")) {

				this.root = null;

			}

			// Racine normale

			else {

				this.root = new Node(root.getName().charAt(0));

				Iterator<Element> i = root.getChildren().iterator();
				int position = 0;

				while (i.hasNext()) {

					recursiveFileConstructor(
							new Address(Integer.toString(position)), i.next());
					position++;

				}

			}

		}

		else
			throw new Exception();

	}

	/**
	 * Constructor building a Tree with an XML document (File) <br />
	 * uses JDOM to parse and build the tree
	 * 
	 * @param file
	 *            File to parse
	 * @param vt
	 *            Vocabulary translator establishing the correspondence between
	 *            XML elements and characters
	 * @throws Exception
	 *             Throws exception if the document isn't valid or if the SAX
	 *             Parser can't parse it or the vocabulary translator is
	 *             overloaded
	 */
	@SuppressWarnings("unchecked")
	public Tree(File file, VocabTranslator vt) throws Exception {

		SAXBuilder sxb = new SAXBuilder();

		org.jdom.Document document;
		document = sxb.build(file);

		Element root = document.getRootElement();

		// Si il y a bien une racine
		if (root != null) {

			// Racine qui d�nome l'arbre vide ?

			if (root.getName().equals("EmptyTree"))
				this.root = null;

			// Racine normale

			else {
				Character root_char;
				// Get the character corresponding to the root label in the
				// vocabulary translator
				try {
					root_char = vt.getChar(root.getName());
				}
				// If no corresponding character then the label does not appear
				// in the DTD
				catch (VocabTranslatorException e) {
					// Add it to the vocabulary translator
					root_char = vt.addElem(root.getName());
				}

				this.root = new Node(root_char);

				Iterator<Element> i = root.getChildren().iterator();
				int position = 0;

				while (i.hasNext()) {

					recursiveFileConstructor(
							new Address(Integer.toString(position)), i.next(),
							vt);
					position++;

				}

			}

		}

		else
			throw new Exception();

	}

	/**
	 * Function used to create resursively for each Element of a Document XML a
	 * Node
	 * 
	 * @param address
	 *            Address where we are
	 * @param element
	 *            Element (JDOM) to add to our Tree
	 * @throws addNodeException
	 *             Throws this Exception if an error occurs in adding the
	 *             Element
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 * @throws headerNodeException
	 *             Throws this exception is the header of the new node is ""
	 */
	@SuppressWarnings("unchecked")
	private void recursiveFileConstructor(Address address, Element element)
			throws addNodeException, badAddressException, headerNodeException {

		addNode(address, element.getName().charAt(0));

		Iterator<Element> i = element.getChildren().iterator();
		int position = 0;

		while (i.hasNext()) {

			recursiveFileConstructor(address.getChildAddress(position),
					i.next());
			position++;

		}

	}

	/**
	 * Function used to create recursively for each Element of a Document XML a
	 * Node
	 * 
	 * @param address
	 *            Address where we are
	 * @param element
	 *            Element (JDOM) to add to our Tree
	 * @param vt
	 *            Vocabulary translator establishing the correspondence between
	 *            XML elements and characters
	 * @throws addNodeException
	 *             Throws this Exception if an error occurs in adding the
	 *             Element
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 * @throws headerNodeException
	 *             Throws this exception is the header of the new node is ""
	 * @throws xmlVocabularySizeTooBig
	 * @throws NullPointerException
	 */
	@SuppressWarnings("unchecked")
	private void recursiveFileConstructor(Address address, Element element,
			VocabTranslator vt) throws addNodeException, badAddressException,
			headerNodeException, NullPointerException, xmlVocabularySizeTooBig {

		// Get the character corresponding to the element's label in the
		// vocabulary translator
		Character elem_char;
		try {
			elem_char = vt.getChar(element.getName());
		}
		// If no character associated to this element (the element is not in the
		// DTD)
		// than add the element to the vocabulary translator
		catch (VocabTranslatorException e) {
			elem_char = vt.addElem(element.getName());
		}

		addNode(address, elem_char);

		Iterator<Element> i = element.getChildren().iterator();
		int position = 0;

		while (i.hasNext()) {

			recursiveFileConstructor(address.getChildAddress(position),
					i.next(), vt);
			position++;

		}

	}

	/**
	 * Returns a copy of the root
	 * 
	 * @return Node root
	 */
	public Node getRoot() {

		return (isEmpty()) ? null : new Node(this.root);

	}

	/**
	 * Returns a direct access to the Tree's root
	 * 
	 * @return Node root
	 */
	private Node getRootReal() {

		return (isEmpty()) ? null : this.root;

	}

	/**
	 * Adds the Node at the Address given<br />
	 * If the Address is already used, it push every other Node to the right<br />
	 * The Node to add can have children
	 * 
	 * @param address
	 *            Address where the new Node should be added
	 * @param node
	 *            Node to add
	 * @throws addNodeException
	 *             throws an addNodeException if the father of the Adresse given
	 *             do not exist
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 */
	private void addNode(Address address, Node node) throws addNodeException,
			badAddressException {

		// Are we trying to add a Node at an empty Tree ?

		if (address.isRoot()) {

			if (!this.isEmpty())
				throw new badAddressException();
			else
				this.root = new Node(node);

		}

		// We are trying to add a normal node

		else {

			Address father_address = address.getFatherAddress();
			Node father;

			// If we try to add a Node to the Root, father_adress should be null
			// else, we can get the father

			father = (father_address.isRoot()) ? getRootReal()
					: getRealNode(father_address);

			// If the father is null, we throw an

			if (father == null)
				throw new addNodeException();

			// Let's insert at the local position
			// Every other Node will be pushed +1 to the right
			// But first, verify that position-1 exists

			int position = address.getPosition();
			if (position > 0
					&& (

					father_address != null
							&& getNode(father_address
									.getChildAddress((position - 1))) == null

					||

					father_address == null
							&& getNode(new Address(String.valueOf(position - 1))) == null

					))
				throw new addNodeException();

			father.addChild(position, node);

		}

	}

	/**
	 * Creates a Node using the string parameter as Header<br />
	 * Add the new Node at the Address given<br />
	 * If the Address is already used, it push every other Node to the right
	 * 
	 * @param address
	 *            Address where the new Node should be added
	 * @param header
	 *            Header to use to create the Node
	 * @throws addNodeException
	 *             throws an addNodeException if the father of the Adresse given
	 *             do not exist
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 * @throws headerNodeException
	 *             Throws this exception is the header of the new node is ""
	 */
	public void addNode(Address address, Character header)
			throws addNodeException, badAddressException, headerNodeException {

		// if(header == "") throw new headerNodeException();
		addNode(address, new Node(header));

	}

	/**
	 * Creates an Adresse using the parameter adress and then uses addNode
	 * 
	 * @param adress
	 *            String to use to create the Address
	 * @param header
	 *            Header to use to create the Node
	 * @throws addNodeException
	 *             throws an addNodeException if the father of the Adresse given
	 *             do not exist
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 * @throws headerNodeException
	 *             Throws this exception is the header of the new node is ""
	 */
	public void addNode(String adress, Character header)
			throws addNodeException, badAddressException, headerNodeException {

		// if(header == "") throw new headerNodeException();
		addNode(new Address(adress), new Node(header));

	}

	/**
	 * Removes the Node at the Address given if it is a Leaf
	 * 
	 * @param address
	 *            Address of the Node to remove
	 * @throws noNodeException
	 *             throws this Exception if no Node exists at the Address given
	 * @throws removeNodeException
	 *             throw this Exception if the Node isn't a Leaf
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 */
	public void removeNode(Address address) throws noNodeException,
			removeNodeException, badAddressException {

		Node node = getNode(address);

		// Does the Node exist ?

		if (node == null)
			throw new noNodeException();

		// Is it a leaf ?

		if (!node.isLeaf())
			throw new removeNodeException();

		// Are we removing the root ?

		if (address.isRoot()) {

			this.root = null;

		}

		// No we arn't

		else {

			// It exists, let's search for the father

			Address father_adress = address.getFatherAddress();
			Node father;

			father = (father_adress.isRoot()) ? getRootReal()
					: getRealNode(father_adress);

			// Let's remove the Node

			int position = address.getPosition();

			father.removeChild(position);

		}

	}

	/**
	 * Creates an Address with the String given and Removes the Node at the
	 * Address if it is a Leaf
	 * 
	 * @param adress
	 *            Address of the Node to remove
	 * @throws noNodeException
	 *             throws this Exception if no Node exists at the Address given
	 * @throws removeNodeException
	 *             throw this Exception if the Node isn't a Leaf
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 * 
	 */
	public void removeNode(String adress) throws noNodeException,
			removeNodeException, badAddressException {

		removeNode(new Address(adress));

	}

	/**
	 * Removes the Node at the Address given even if it isn't a Leaf
	 * 
	 * @param address
	 *            Address of the Node to remove
	 * @throws noNodeException
	 *             throws this Exception if no Node exists at the Address given
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 * 
	 */
	public void removeTree(Address address) throws noNodeException,
			badAddressException {

		// Do we try to remove the root ?

		if (address.isRoot()) {

			if (!this.isEmpty()) {

				this.root = null;

			} else
				throw new noNodeException();

		}

		// No, we remove a sub-tree

		else {

			// Does the Node exist ?

			if (getNode(address) == null)
				throw new noNodeException();

			// It exists, let's search for the father

			Address father_adress = address.getFatherAddress();
			Node father;

			father = (father_adress.isRoot()) ? getRootReal()
					: getRealNode(father_adress);

			// Let's remove the Node

			int position = address.getPosition();

			father.removeChild(position);

		}

	}

	/**
	 * Creates an Address with the String given and removes the Node at this
	 * Address even if it isn't a Leaf
	 * 
	 * @param address
	 *            Address of the Node to remove
	 * @throws noNodeException
	 *             throws this Exception if no Node exists at the Address given
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 * 
	 */
	public void removeTree(String address) throws noNodeException,
			badAddressException {

		removeTree(new Address(address));

	}

	/**
	 * Adds the Tree at the Address given
	 * 
	 * @param address
	 *            Address where the tree should be added
	 * @param tree
	 *            Tree to add
	 * @throws addNodeException
	 *             Throws this exception if the insertion can't be made (father
	 *             doesn't exists...)
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 */
	public void addTree(Address address, Tree tree) throws addNodeException,
			badAddressException {

		if (!tree.isEmpty())
			addNode(address, tree.getRoot());

	}

	/**
	 * Creates an Address with the String given and adds the Tree at this
	 * Address
	 * 
	 * @param address
	 *            Address where the tree should be added
	 * @param tree
	 *            Tree to add
	 * @throws addNodeException
	 *             Throws this exception if the insertion can't be made (father
	 *             doesn't exists...)
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 */
	public void addTree(String address, Tree tree) throws addNodeException,
			badAddressException {

		if (!tree.isEmpty())
			addNode(new Address(address), tree.getRoot());

	}

	/**
	 * Changes the Label (Header) of the Node at the given Address with the
	 * given Header
	 * 
	 * @param address
	 *            Address of the Node to relabel
	 * @param header
	 *            Header to use to relabel the node
	 * @throws noNodeException
	 *             Throws this Exception if there isn't any Node at the given
	 *             Address
	 * @throws headerNodeException
	 *             If the Header given is empty
	 */
	public void relabelNode(Address address, Character header)
			throws noNodeException, headerNodeException {

		// Is the label valid ?
		// if(header == "") throw new headerNodeException();

		Node node = getRealNode(address);

		// Does the node exist ?

		if (node == null)
			throw new noNodeException();

		// It exists, lets change it's header

		node.setHeader(header);

	}

	/**
	 * Creates an Address using the String address given and Changes the Label
	 * (Header) of the Node at the Address with the given Header
	 * 
	 * @param address
	 *            String of the Address of the Node to relabel
	 * @param header
	 *            Header to use to relabel the node
	 * @throws noNodeException
	 *             Throws this Exception if there isn't any Node at the given
	 *             Address
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 * @throws headerNodeException
	 *             If the Header given is empty
	 */
	public void relabelNode(String address, Character header)
			throws noNodeException, badAddressException, headerNodeException {

		// if(header == "") throw new headerNodeException();
		relabelNode(new Address(address), header);

	}

	/**
	 * Replace the Node and its children at the given Address with the tree
	 * given
	 * 
	 * @param address
	 *            Address of the Node that should be replaced
	 * @param tree
	 *            Tree that has to replace the Node
	 * @throws noNodeException
	 *             Throws this Exception if there is no Node at the adress Given
	 * @throws addNodeException
	 *             Throws this Exception if the Tree cannot be added
	 * @throws removeNodeException
	 *             Throws this Exception if the Node at the adress given cannot
	 *             be removed
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 */
	public void replaceTree(Address address, Tree tree) throws noNodeException,
			addNodeException, removeNodeException, badAddressException {

		// Is it a root replace ?

		if (address.isRoot()) {

			this.root = tree.getRoot();

		}

		// It's not

		else {

			// Let's remove the tree

			removeTree(address);

			// And add a new one

			addTree(address, tree);
		}

	}

	/**
	 * Creates an Address with the String given and replace the Node and its
	 * children at this Address with the tree given
	 * 
	 * @param address
	 *            Address of the Node that should be replaced
	 * @param tree
	 *            Tree that has to replace the Node
	 * @throws noNodeException
	 *             Throws this Exception if there is no Node at the address
	 *             Given
	 * @throws addNodeException
	 *             Throws this Exception if the Tree cannot be added
	 * @throws removeNodeException
	 *             Throws this Exception if the Node at the address given cannot
	 *             be removed
	 * @throws badAddressException
	 *             Throws this Exception if there is an error while creating new
	 *             Addresses
	 */
	public void replaceTree(String address, Tree tree) throws noNodeException,
			addNodeException, removeNodeException, badAddressException {

		replaceTree(new Address(address), tree);

	}

	/**
	 * Returns the direct access to the Node at the given adress
	 * 
	 * @param address
	 *            Address of the Node to return
	 * @return the Node or null if it doesn't exists
	 */
	private Node getRealNode(Address address) {

		if (address.isRoot())
			return getRootReal();

		else {

			Node resultat = getRootReal();
			String[] adresses = address.getPositions();

			for (int i = 0; i < adresses.length; i++) {

				if (resultat == null)
					break;
				else
					resultat = resultat.getChild(Integer.valueOf(adresses[i]));

			}

			return resultat;

		}

	}

	/**
	 * Returns the cost of the creation of the tree using the cost
	 * AddNodeOperation.cost for every leaf
	 * 
	 * @return Cost of the creation
	 */
	public int getCreationCost() {

		return (isEmpty()) ? 0 : getCreationCostRecursive(getRoot());

	}

	/**
	 * Function used by getCreationCost to calculate the cost of creating the
	 * tree
	 * 
	 * @param node
	 * @return the cost
	 */
	private int getCreationCostRecursive(Node node) {

		int cost = AddNodeOperation.cost;
		int i = 0;

		while (node.getChild(i) != null) {

			cost = cost + getCreationCostRecursive(node.getChild(i));
			i++;

		}

		return cost;

	}

	/**
	 * Returns the cost of the removal of the tree using the cost
	 * RemoveNodeOperation.cost for every leaf
	 * 
	 * @return Cost of the removal
	 */
	public int getRemovalCost() {

		Node root = getRoot();

		return getRemovalCostRecursive(root);

	}

	/**
	 * Function used by getRemovalCost to get the cost of the remove of a tree
	 * 
	 * @param node
	 * @return the cost
	 */
	private int getRemovalCostRecursive(Node node) {

		int cost = RemoveNodeOperation.cost;
		int i = 0;

		while (node.getChild(i) != null) {

			cost = cost + getRemovalCostRecursive(node.getChild(i));
			i++;

		}

		return cost;

	}

	/**
	 * Returns a copy of the Node at the given address
	 * 
	 * @param address
	 *            Address of the Node
	 * @return A copy of the Node or null if it doesn't exists
	 */
	public Node getNode(Address address) {

		if (address.isRoot())
			return getRoot();

		else {

			Node resultat = getRootReal();
			String[] adresses = address.getPositions();

			for (int i = 0; i < adresses.length; i++) {

				if (resultat == null)
					break;
				else
					resultat = resultat.getChild(Integer.valueOf(adresses[i]));

			}

			return (resultat == null) ? resultat : new Node(resultat);

		}

	}

	/**
	 * Returns Tree create from the Node at the address given
	 * 
	 * @param address
	 *            Address of the Node
	 * @return A Tree
	 * @throws noNodeException
	 *             Throws this exception if there isn't any Node at the address
	 *             given
	 */
	public Tree getNodeAsTree(Address address) throws noNodeException {

		Node n = getNode(address);

		if (n == null)
			throw new noNodeException();

		return new Tree(n);

	}

	/**
	 * Returns true if the root is empty
	 * 
	 * @return true if the root doesn't exit
	 */
	public boolean isEmpty() {

		return this.root == null;

	}

	/**
	 * M�thode utilis�e par la m�thode toString afin d'afficher de fa�on claire
	 * l'ensemble des fils de l'arbre
	 * 
	 * @param address
	 *            adresse courrant des noeuds � afficher
	 * @return une cha�ne de caract�re
	 */
	private String recursiveToFlatString(Address address) {

		String resultat = "";
		int i = 0;
		Node n;

		try {

			while (getNode(address.getChildAddress(i)) != null) {

				n = getNode(address.getChildAddress(i));

				if (getNode(address.getChildAddress(i).getChildAddress(0)) == null) {

					resultat = resultat + "<" + n + " />";

				} else {

					resultat = resultat + "<" + n + ">";
					resultat = resultat
							+ recursiveToFlatString(address.getChildAddress(i));
					resultat = resultat + "</" + n + ">";

				}

				i++;

			}

		} catch (badAddressException e) {

			resultat = "Error while creating Addresses";

		}

		return resultat;

	}

	/**
	 * M�thode utilis�e par la m�thode toString afin d'afficher de fa�on claire
	 * l'ensemble des fils de l'arbre
	 * 
	 * @param address
	 *            adresse courrant des noeuds � afficher
	 * 
	 * @param vt
	 *            Vocabulary translator establishing the correspondence between
	 *            XML elements and characters
	 * 
	 * @return une cha�ne de caract�re
	 */
	private String recursiveToFlatString(Address address, VocabTranslator vt) {

		String resultat = "";
		int i = 0;
		Node n;

		try {

			while (getNode(address.getChildAddress(i)) != null) {

				n = getNode(address.getChildAddress(i));

				if (getNode(address.getChildAddress(i).getChildAddress(0)) == null) {

					resultat = resultat + "<" + vt.getElem(n.getHeader())
							+ " />";

				} else {

					resultat = resultat + "<" + vt.getElem(n.getHeader()) + ">";
					resultat = resultat
							+ recursiveToFlatString(address.getChildAddress(i),
									vt);
					resultat = resultat + "</" + vt.getElem(n.getHeader())
							+ ">";

				}

				i++;

			}

		} catch (badAddressException e) {

			resultat = "Error while creating Addresses";

		}

		return resultat;

	}

	/**
	 * Let the user print the Tree this way: <br />
	 * &lt;root&gt;&lt;node /&gt;&lt;/root&gt;
	 * 
	 * @return a String
	 */
	public String toFlatString() {

		String resultat = "";

		// Is the tree empty ?

		if (this.isEmpty()) {

			resultat = "<empty tree>";

		}

		// Leaf ?

		else if (getRoot().isLeaf()) {

			resultat = "<" + getRootReal().getHeader() + " />";

		}
		// It isn't !

		else {

			resultat = resultat + "<" + getRootReal().getHeader() + ">";

			int i = 0;
			Node n = null;

			try {

				while (getNode(new Address(Integer.toString(i))) != null) {

					n = getNode(new Address(String.valueOf(i)));

					if (getNode(new Address(String.valueOf(i))
							.getChildAddress(0)) == null) {

						resultat = resultat + "<" + n + " />";

					}

					else {

						resultat = resultat + "<" + n + ">";
						resultat = resultat
								+ recursiveToFlatString(new Address(
										String.valueOf(i)));
						resultat = resultat + "</" + n + ">";

					}

					i++;

				}

				resultat = resultat + "</" + getRootReal().getHeader() + ">";

			} catch (badAddressException e) {

				resultat = "Error while creating Addresses";

			}

		}

		return resultat;

	}

	/**
	 * Let the user print the Tree this way: <br />
	 * &lt;root&gt;&lt;node /&gt;&lt;/root&gt;
	 * 
	 * @param vt
	 *            Vocabulary translator establishing the correspondence between
	 *            XML elements and characters
	 * 
	 * @return a String
	 */
	public String toFlatString(VocabTranslator vt) {

		String resultat = "";

		// Is the tree empty ?

		if (this.isEmpty()) {

			resultat = "<empty tree>";

		}

		// Leaf ?

		else if (getRoot().isLeaf()) {

			resultat = "<" + vt.getElem(getRootReal().getHeader()) + " />";

		}
		// It isn't !

		else {

			resultat = resultat + "<" + vt.getElem(getRootReal().getHeader())
					+ ">";

			int i = 0;
			Node n = null;

			try {

				while (getNode(new Address(Integer.toString(i))) != null) {

					n = getNode(new Address(String.valueOf(i)));

					if (getNode(new Address(String.valueOf(i))
							.getChildAddress(0)) == null) {

						resultat = resultat + "<" + vt.getElem(n.getHeader())
								+ " />";

					}

					else {

						resultat = resultat + "<" + vt.getElem(n.getHeader())
								+ ">";
						resultat = resultat
								+ recursiveToFlatString(
										new Address(String.valueOf(i)), vt);
						resultat = resultat + "</" + vt.getElem(n.getHeader())
								+ ">";

					}

					i++;

				}

				resultat = resultat + "</"
						+ vt.getElem(getRootReal().getHeader()) + ">";

			} catch (badAddressException e) {

				resultat = "Error while creating Addresses";

			}

		}

		return resultat;

	}

	/**
	 * M�thode utilis�e par la m�thode toString afin d'afficher de fa�on claire
	 * l'ensemble des fils de l'arbre
	 * 
	 * @param address
	 *            adresse courrant des noeuds � afficher
	 * @param tab
	 *            tabulation permettant un affichage propre
	 * @return une cha�ne de caract�re
	 */
	private String recursiveToString(Address address, String tab) {

		String resultat = "";
		int i = 0;
		Node n;

		try {

			while (getNode(address.getChildAddress(i)) != null) {

				n = getNode(address.getChildAddress(i));

				if (getNode(address.getChildAddress(i).getChildAddress(0)) == null) {

					resultat = resultat + tab + "<" + n + ":"
							+ address.getChildAddress(i) + " />\n";

				} else {

					resultat = resultat + tab + "<" + n + ":"
							+ address.getChildAddress(i) + ">\n";
					resultat = resultat
							+ recursiveToString(address.getChildAddress(i), tab
									+ "\t");
					resultat = resultat + tab + "</" + n + ">\n";

				}

				i++;

			}

		} catch (badAddressException e) {

			resultat = "Error while creating Addresses";

		}

		return resultat;

	}

	/**
	 * M�thode utilis�e par la m�thode toString afin d'afficher de fa�on claire
	 * l'ensemble des fils de l'arbre
	 * 
	 * @param address
	 *            adresse courrant des noeuds � afficher
	 * @param tab
	 *            tabulation permettant un affichage propre
	 * 
	 * @param vt
	 *            Vocabulary translator establishing the correspondence between
	 *            XML elements and characters
	 * 
	 * @return une cha�ne de caract�re
	 */
	private String recursiveToString(Address address, String tab,
			VocabTranslator vt) {

		String resultat = "";
		int i = 0;
		Node n;

		try {

			while (getNode(address.getChildAddress(i)) != null) {

				n = getNode(address.getChildAddress(i));

				if (getNode(address.getChildAddress(i).getChildAddress(0)) == null) {

					resultat = resultat + tab + "<" + vt.getElem(n.getHeader())
							+ ":" + address.getChildAddress(i) + " />\n";

				} else {

					resultat = resultat + tab + "<" + vt.getElem(n.getHeader())
							+ ":" + address.getChildAddress(i) + ">\n";
					resultat = resultat
							+ recursiveToString(address.getChildAddress(i), tab
									+ "\t", vt);
					resultat = resultat + tab + "</"
							+ vt.getElem(n.getHeader()) + ">\n";

				}

				i++;

			}

		} catch (badAddressException e) {

			resultat = "Error while creating Addresses";

		}

		return resultat;

	}

	/**
	 * Let the user print the Tree this way: <br />
	 * &lt;root&gt; <br />
	 * &lt;node /&gt; <br />
	 * &lt;/root&gt;
	 * 
	 * @return a String
	 */
	public String toString() {

		String resultat = "";

		// Is the tree empty ?

		if (this.isEmpty()) {

			resultat = "<empty tree>";

		}

		// It isn't !

		else {

			resultat = resultat + "<" + getRootReal().getHeader() + ">\n";

			int i = 0;
			Node n = null;

			try {

				while (getNode(new Address(Integer.toString(i))) != null) {

					n = getNode(new Address(String.valueOf(i)));

					if (getNode(new Address(String.valueOf(i))
							.getChildAddress(0)) == null) {

						resultat = resultat + "\t<" + n + ":"
								+ (new Address(String.valueOf(i))) + " />\n";

					}

					else {

						resultat = resultat + "\t<" + n + ":"
								+ (new Address(String.valueOf(i))) + ">\n";
						resultat = resultat
								+ recursiveToString(
										new Address(String.valueOf(i)), "\t\t");
						resultat = resultat + "\t</" + n + ">\n";

					}

					i++;

				}

				resultat = resultat + "</" + getRootReal().getHeader() + ">\n";

			} catch (badAddressException e) {

				resultat = "Error while creating Addresses";

			}

		}

		return resultat;

	}

	/**
	 * Let the user print the Tree this way: <br />
	 * &lt;root&gt; <br />
	 * &lt;node /&gt; <br />
	 * &lt;/root&gt;
	 * 
	 * @param vt
	 *            Vocabulary translator establishing the correspondence between
	 *            XML elements and characters
	 * 
	 * @return a String
	 */
	public String toString(VocabTranslator vt) {

		String resultat = "";

		// Is the tree empty ?

		if (this.isEmpty()) {

			resultat = "<empty tree>";

		}

		// It isn't !

		else {

			resultat = resultat + "<" + vt.getElem(getRootReal().getHeader())
					+ ">\n";

			int i = 0;
			Node n = null;

			try {

				while (getNode(new Address(Integer.toString(i))) != null) {

					n = getNode(new Address(String.valueOf(i)));

					if (getNode(new Address(String.valueOf(i))
							.getChildAddress(0)) == null) {

						resultat = resultat + "\t<" + vt.getElem(n.getHeader())
								+ ":" + (new Address(String.valueOf(i)))
								+ " />\n";

					}

					else {

						resultat = resultat + "\t<" + vt.getElem(n.getHeader())
								+ ":" + (new Address(String.valueOf(i)))
								+ ">\n";
						resultat = resultat
								+ recursiveToString(
										new Address(String.valueOf(i)), "\t\t",
										vt);
						resultat = resultat + "\t</"
								+ vt.getElem(n.getHeader()) + ">\n";

					}

					i++;

				}

				resultat = resultat + "</"
						+ vt.getElem(getRootReal().getHeader()) + ">\n";

			} catch (badAddressException e) {

				resultat = "Error while creating Addresses";

			}

		}

		return resultat;

	}

	/**
	 * Function used by getCreationOperationsSequence to create the
	 * OperationSequence to create the tree
	 * 
	 * @param address
	 * @param os
	 * @throws badOperationException
	 * @throws badAddressException
	 * @throws headerNodeException
	 */
	private void getCreationOperationsSequenceRecursive(Address address,
			OperationSequence os) throws badOperationException,
			badAddressException, headerNodeException {

		int i = 0;
		Node n = getNode(address.getChildAddress(i));

		// Let's go while there's a child
		while (n != null) {

			// Add child's creation
			os.addOperation(new AddNodeOperation(address.getChildAddress(i), n
					.getHeader()));

			// If he's not a leaf, go recursive
			if (!n.isLeaf())
				this.getCreationOperationsSequenceRecursive(
						address.getChildAddress(i), os);

			i++;
			n = getNode(address.getChildAddress(i));

		}

	}

	/**
	 * Returns an OperationSequence that can be used to create the current tree
	 * from an empty tree
	 * 
	 * @return OperationsSquence to make
	 * @throws badOperationException
	 *             Throws this Exception if there is an error while creating the
	 *             operations
	 * @throws badAddressException
	 *             Throws this Exception if there is an error with the addresses
	 * @throws headerNodeException
	 *             Throws this exception is the header of the new node is ""
	 */
	public OperationSequence getCreationOperationsSequence()
			throws badOperationException, badAddressException,
			headerNodeException {

		// Let's create a new sequence
		OperationSequence os = new OperationSequence();

		// if our tree is empty, nothing to do
		if (this.isEmpty())
			return os;

		// add root's creation
		os.addOperation(new AddNodeOperation(new Address(""), root.getHeader()));

		// recursive for every sub tree of the root
		getCreationOperationsSequenceRecursive(new Address(""), os);

		return os;

	}

	/**
	 * Function used by getRemoveOperationsSequence to create the
	 * OperationSequence to remove a Tree
	 * 
	 * @param address
	 * @param os
	 * @throws badOperationException
	 * @throws badAddressException
	 */
	private void getRemoveOperationsSequenceRecursive(Address address,
			OperationSequence os) throws badOperationException,
			badAddressException {

		// Let's get the current Node to remove and its children number
		Node n = getNode(address);
		int i = n.getChildrenNumber() - 1;

		// for every children
		for (; i >= 0; i--) {

			// Get the child
			n = getNode(address.getChildAddress(i));

			// Go recursive
			if (!n.isLeaf())
				this.getRemoveOperationsSequenceRecursive(
						address.getChildAddress(i), os);

			// Add the child remove
			os.addOperation(new RemoveNodeOperation(address.getChildAddress(i)));

		}

	}

	/**
	 * Returns an OperationSequence that can be used to remove completely the
	 * current tree
	 * 
	 * @return OperationsSquence to make
	 * @throws badOperationException
	 *             Throws this Exception if there is an error while creating the
	 *             operations
	 * @throws badAddressException
	 *             Throws this Exception if there is an error with the addresses
	 */

	public OperationSequence getRemoveOperationsSequence()
			throws badOperationException, badAddressException {

		// let's create a new seq
		OperationSequence os = new OperationSequence();

		// if our tree is empty, nothing to do
		if (this.isEmpty())
			return os;

		// recursive for every sub tree of the root
		this.getRemoveOperationsSequenceRecursive(new Address(""), os);

		// add root's remove
		os.addOperation(new RemoveNodeOperation(new Address("")));

		return os;

	}

	public int hashCode() {

		return this.toFlatString().hashCode();

	}

	public boolean equals(Object o) {

		if (o instanceof Tree)
			return o.hashCode() == this.hashCode();
		else
			return false;

	}

	public Tree clone() {

		return new Tree(this);

	}

}
