/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.vocabtranslator;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

import fr.ufrt.xmlcorrector.exceptions.*;

/**
 * Class BiHashtable.
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class BiHashtable {

	// A mapping from strings to single characters
	private Hashtable<String, Character> strings2chars;

	// A mapping from single characters to strings
	private Hashtable<Character, String> chars2strings;

	/**
	 * Initialises the bi-directional hashtable with the vocabulary size
	 * 
	 * @param nbStrings
	 *            number of strings in the vocabulary
	 */
	public BiHashtable(int nbStrings) {
		strings2chars = new Hashtable<String, Character>(nbStrings);
		chars2strings = new Hashtable<Character, String>(nbStrings);
	}

	/**
	 * Initialises the bi-directional hashtable translator with the vocabulary
	 * list
	 * 
	 * @param vocab
	 *            List of words to be translated. The corresponding letters will
	 *            start with ascii 'a'.
	 * @throws xmlVocabularySizeTooBig
	 *             if no more association can be added
	 */
	public BiHashtable(List<String> vocab) throws xmlVocabularySizeTooBig {
		if (vocab != null) {
			Integer nbStrings = vocab.size();
			strings2chars = new Hashtable<String, Character>(nbStrings);
			chars2strings = new Hashtable<Character, String>(nbStrings);

			Iterator<String> i = vocab.iterator();
			while (i.hasNext())
				addString(i.next());
		}
	}

	/**
	 * Get the character associated to the string
	 * 
	 * @param str
	 *            string
	 * @throws NullPointerException
	 *             if the element is null
	 * @return the character associated with the string or null if str is
	 *         unknown
	 */
	public Character getChar(String str) throws NullPointerException {
		return strings2chars.get(str);
	}

	/**
	 * Returns the string associated to the character
	 * 
	 * @param c
	 *            character for which the string is needed
	 * @throws NullPointerException
	 *             if the element is null
	 * @return the string associated to the character or null if c is unknown
	 */
	public String getString(Character c) throws NullPointerException {
		return chars2strings.get(c);
	}

	/**
	 * @return Establish the mutual association between the string and the
	 *         character
	 * @param str
	 *            string to be associated
	 * @param c
	 *            character to be associated
	 * @throws NullPointerException
	 *             if the string or the character is null
	 */
	public void associate(String str, Character c) throws NullPointerException {
		this.strings2chars.put(str, c);
		this.chars2strings.put(c, str);
	}

	/**
	 * @return Establish the mutual association between the element and the
	 *         character
	 * @param elem
	 *            element to be associated
	 * @param c
	 *            character to be associated
	 * @throws NullPointerException
	 *             if the element or the character is null
	 * @throws xmlVocabularySizeTooBig
	 *             if no more association can be added
	 */
	public Character addString(String str) throws NullPointerException,
			xmlVocabularySizeTooBig {
		// With this method the highest possible character is the last visible
		// ASCII character
		if (this.strings2chars.size() + (int) 'a' >= (int) '~')
			throw new xmlVocabularySizeTooBig();
		// new character to be associated to the element
		Character newChar = new Character(
				(char) (this.strings2chars.size() + (int) 'a'));
		this.strings2chars.put(str, newChar);
		this.chars2strings.put(newChar, str);
		return newChar;
	}

	/**
	 * Get the string representation of the word-to-character equivalences
	 * contained in the translator
	 * 
	 * @throws NullPointerException
	 *             if the element or the character is null
	 */
	public String toString() {
		return strings2chars.toString();
	}

}
