/*
 *   Copyright 2012 Joshua AMAVI, Alexandre BOREL, Béatrice BOUCHOU, Agata SAVARY
 *   <joshua.amavi@gmail.com><beatrice.bouchou@univ-tours.fr><agata.savary@univ-tours.fr>
 *
 *   This file is part of XMLCorrector.
 *
 *   XMLCorrector is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   XMLCorrector is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License,
 *   along with XMLCorrector.  If not, see <http://www.gnu.org/licenses/>.
*/

package fr.ufrt.xmlcorrector.vocabtranslator;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Enumeration;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

//import com.sun.org.apache.xerces.internal.parsers.DTDParser;

import fr.ufrt.xmlcorrector.exceptions.VocabTranslatorException;
import fr.ufrt.xmlcorrector.exceptions.anyNotTreatedInDTDException;
import fr.ufrt.xmlcorrector.exceptions.xmlVocabularySizeTooBig;

import com.wutka.dtd.*;

/**
 * Class VocabTranslator: Encoding string to character
 * 
 * @author Joshua AMAVI
 * @author Alexandre BOREL
 * @author Béatrice BOUCHOU
 * @author Agata SAVARY
 */
public class VocabTranslator {

	// Bi-directional hastable establishing correspondence between string DTD
	// elements and characters
	private BiHashtable vt = new BiHashtable(0);

	public VocabTranslator() {
	}

	/**
	 * Get the character associated to the XML element
	 * 
	 * @param elem
	 *            XML element
	 * @throws NullPointerException
	 *             if the element is null
	 * @return the character associated with the string or null if elem is
	 *         unknown
	 */
	public char getChar(String elem) throws NullPointerException,
			VocabTranslatorException {
		if (vt.getChar(elem) == null)
			throw new VocabTranslatorException();
		else
			// return (char) vt.getChar(elem);
			return vt.getChar(elem);
	}

	/**
	 * Returns the XML element associated to the character
	 * 
	 * @param c
	 *            character for which the XML element is needed
	 * @throws NullPointerException
	 *             if the element is null
	 * @return the XML element associated to the character or null if c is
	 *         unknown
	 */
	public String getElem(char c) throws NullPointerException {
		Character ch = new Character(c);
		return vt.getString(ch);
	}

	/**
	 * Adds a new XML element to the translator
	 * 
	 * @return Character associated to elem
	 * @param elem
	 *            element to be associated character to be associated
	 * @throws NullPointerException
	 *             if the element or the character is null
	 * @throws xmlVocabularySizeTooBig
	 *             if no more association can be added
	 */
	public char addElem(String elem) throws NullPointerException,
			xmlVocabularySizeTooBig {
		return (char) vt.addString(elem);
	}

	/**
	 * Create and XML file from a DTD by keeping only element names and
	 * translating words into characters (attributes, PCDATA and "Any" are not
	 * treated) E.g. the following DTD: <!ELEMENT donneeVente
	 * (client*,unite*,total)> <!ELEMENT client EMPTY> <!ELEMENT unite EMPTY>
	 * <!ELEMENT total (totalClient+,(totalPanier|totalUnite)*)> <!ELEMENT
	 * totalClient EMPTY> <!ELEMENT totalPanier EMPTY> <!ELEMENT totalUnite
	 * EMPTY> is translated into the following XML file: <?xml version="1.0"
	 * encoding="UTF-8"?> <TreeSchema> <root>a</root> <!-- donneeVente --> <rule
	 * node="a" source="donneeVente (client*,unite*,total)">b*c*d</rule> <rule
	 * node="b" source="client EMPTY"></rule> <rule node="c"
	 * source="unite EMPTY"></rule> <rule node="d"
	 * source="total (totalClient+,(totalPanier|totalUnite)*)">e+(f|g)*</rule>
	 * <rule node="e" source="totalClient EMPTY"></rule> <rule node="f"
	 * source="totalPanier EMPTY"></rule> <rule node="g"
	 * source="otalUnite EMPTY"></rule> </TreeSchema>
	 * 
	 * @param dtdFile
	 *            DTD file to be transformed
	 * @param dtdXmlFile
	 *            XML file containing the translated DTD
	 * @throws anyNotTreatedInDTDException
	 *             if a DTD element's content is of "Any" type
	 * @throws IOException
	 * @throws MalformedURLException
	 * @throws xmlVocabularySizeTooBig
	 * @throws NullPointerException
	 */
	public void DTDtoXML(String dtdFile, String dtdXmlFile)
			throws anyNotTreatedInDTDException, MalformedURLException,
			IOException, NullPointerException, xmlVocabularySizeTooBig {
		// try {
		DTDParser parser = null;
		// MAW Version 1.17
		// If it looks like the filename may be a URL, use the URL class
		if (dtdFile.indexOf("://") > 0)
			parser = new DTDParser(new URL(dtdFile), false);
		else
			parser = new DTDParser(new File(dtdFile), false);
		// Parse the DTD and ask the parser to guess the root element
		DTD dtd = parser.parse(true);

		Enumeration<DTDElement> dtdelems = dtd.elements.elements();
		while (dtdelems.hasMoreElements()) {
			DTDElement elem = (DTDElement) dtdelems.nextElement();
			vt.addString(elem.getName());
		}
		// System.out.println("Vocab translator: \n" + vt.toString());

		Document xmldtd = translateDTD(dtd);
		XMLOutputter outputter = new XMLOutputter(Format.getPrettyFormat());
		outputter.output(xmldtd, new FileOutputStream(dtdXmlFile));
		// }
		// catch (Exception exc) {
		// System.out.println("Problem occured while translating the DTD into an XML one-letter DTD.\n");
		// exc.printStackTrace(System.out);
		// }
	}

	/**
	 * transform a DTD to an XML document while translating words into
	 * characters (attributes, PCDATA and "Any" are not treated) E.g. the
	 * following DTD: <!ELEMENT donneeVente (client*,unite*,total)> <!ELEMENT
	 * client EMPTY> <!ELEMENT unite EMPTY> <!ELEMENT total
	 * (totalClient+,(totalPanier|totalUnite)*)> <!ELEMENT totalClient EMPTY>
	 * <!ELEMENT totalPanier EMPTY> <!ELEMENT totalUnite EMPTY> is translated
	 * into the following XML document: <TreeSchema> <root>a</root> <!--
	 * donneeVente --> <rule node="a"
	 * source="donneeVente (client*,unite*,total)">b*c*d</rule> <rule node="b"
	 * source="client EMPTY"></rule> <rule node="c" source="unite EMPTY"></rule>
	 * <rule node="d"
	 * source="total (totalClient+,(totalPanier|totalUnite)*)">e+(f|g)*</rule>
	 * <rule node="e" source="totalClient EMPTY"></rule> <rule node="f"
	 * source="totalPanier EMPTY"></rule> <rule node="g"
	 * source="otalUnite EMPTY"></rule> </TreeSchema>
	 * 
	 * @param dtd
	 *            DTD to be transformed
	 * @return XML document representing the DTD e.g.
	 * @throws anyNotTreatedInDTDException
	 *             if a DTD element's content is of "Any" type
	 */
	public Document translateDTD(DTD dtd) throws anyNotTreatedInDTDException {

		// Translate the root
		if (dtd.rootElement == null)
			return null;
		Element root = new Element("TreeSchema"); // Root of the target XML
													// document
		Element dtdroot = new Element("root"); // Root of the DTD
		// Translated the root's name into a character
		dtdroot.addContent(vt.getChar(dtd.rootElement.getName()).toString());
		root.addContent(dtdroot);
		Document xmldtd = new Document(root); // Target XML document
												// representing the DTD

		// Translate root's children
		Enumeration<DTDElement> dtdelems = dtd.elements.elements();
		while (dtdelems.hasMoreElements()) {
			DTDElement elem = (DTDElement) dtdelems.nextElement();
			Element xmlelem = translateDTDElement(elem);
			root.addContent(xmlelem);
			// xmldtd.createComment(elem.name + " " + elem.content);
		}
		return xmldtd;
	}

	/**
	 * transform a DTD element to an XML element by translating words into
	 * characters (attributes, PCDATA and "Any" are not treated)
	 * 
	 * @param item
	 *            DTD element to be transformed e.g. total
	 *            (totalClient+,(totalPanier|totalUnite)*)
	 * @return XML Element representing the DTD element e.g. <rule
	 *         node="d">e+(f|g)*</rule>
	 * @throws anyNotTreatedInDTDException
	 *             if the DTD element is of "Any" type
	 */
	public Element translateDTDElement(DTDElement elem)
			throws anyNotTreatedInDTDException {
		// System.out.println("Element do przetlumaczenia: " + elem.name + " " +
		// elem.content.toString());

		Element rule = new Element("rule");
		rule.setAttribute("node", vt.getChar(elem.name).toString());
		rule.setAttribute("source", elem.name + " " + elem.content.toString());
		rule.addContent(translateDTDItem(elem.content));
		return rule;
	}

	/**
	 * transform a DTD element's content to a string while translating it into a
	 * character (attributes, PCDATA and "Any" are not treated)
	 * 
	 * @param item
	 *            DTD element's content to be transformed e.g.
	 *            (totalClient*,(totalPanier|totalUnite)*)
	 * @return string representing the DTD element e.g. b*(c|d)*
	 * @throws anyNotTreatedInDTDException
	 *             if the DTD element is of "Any" type
	 */
	public String translateDTDItem(DTDItem item)
			throws anyNotTreatedInDTDException {

		if (item == null)
			return null;

		if (item instanceof DTDAny) // Any-type elements are not treated by our
									// corrector (they would generate infinite
									// number of solutions)
			throw new anyNotTreatedInDTDException();
		else if (item instanceof DTDEmpty || item instanceof DTDPCData)
			return "";
		else {
			String result = ""; // Result string representing the translated
								// element's content
			if (item instanceof DTDName) {// Element name
				// Translate string element to a char
				result = vt.getChar(((DTDName) item).value).toString();
			} else {
				result = "(";
				DTDItem[] items = null;
				if (item instanceof DTDChoice)
					items = ((DTDChoice) item).getItems();
				else if (item instanceof DTDSequence)
					items = ((DTDSequence) item).getItems();
				else if (item instanceof DTDMixed)
					items = ((DTDMixed) item).getItems();
				for (int i = 0; i < items.length; i++) {
					if (i > 0 && item instanceof DTDChoice)
						result = result + "|";
					result = result + translateDTDItem(items[i]);
				}
				result = result + ")";
			}

			// Translate the cardinality *, + or ?
			if (item.cardinal == DTDCardinal.OPTIONAL)
				result = result + "?";
			else if (item.cardinal == DTDCardinal.ZEROMANY)
				result = result + "*";
			else if (item.cardinal == DTDCardinal.ONEMANY) // OK
				result = result + "+";
			return result;
		}
	}
}
